@@ -15,40 +15,26 @@
 
 import webob
 
+from oslo_log import log as logging
+from oslo_utils import strutils
+
 from cinder.api import extensions
 from cinder.api.openstack import wsgi
-from cinder.api import xmlutil
+from cinder.api.schemas import quotas
+from cinder.api import validation
 from cinder import db
-from cinder.db.sqlalchemy import api as sqlalchemy_api
 from cinder import exception
 from cinder.i18n import _
+from cinder.policies import quotas as policy
 from cinder import quota
 from cinder import quota_utils
 from cinder import utils
 
-from oslo_config import cfg
-from oslo_utils import strutils
+LOG = logging.getLogger(__name__)
 
-
-CONF = cfg.CONF
 QUOTAS = quota.QUOTAS
-NON_QUOTA_KEYS = ['tenant_id', 'id']
-
-authorize_update = extensions.extension_authorizer('volume', 'quotas:update')
-authorize_show = extensions.extension_authorizer('volume', 'quotas:show')
-authorize_delete = extensions.extension_authorizer('volume', 'quotas:delete')
-
-
-class QuotaTemplate(xmlutil.TemplateBuilder):
-    def construct(self):
-        root = xmlutil.TemplateElement('quota_set', selector='quota_set')
-        root.set('id')
-
-        for resource in QUOTAS.resources:
-            elem = xmlutil.SubTemplateElement(root, resource)
-            elem.text = resource
-
-        return xmlutil.MasterTemplate(root, 1)
+GROUP_QUOTAS = quota.GROUP_QUOTAS
+NON_QUOTA_KEYS = quota.NON_QUOTA_KEYS
 
 
 class QuotaSetsController(wsgi.Controller):
@@ -69,14 +55,19 @@ class QuotaSetsController(wsgi.Controller):
         if QUOTAS.using_nested_quotas():
             used += v.get('allocated', 0)
         if value < used:
-            # TODO(mc_nair): after N opens, update error message to include
-            # the current usage and requested limit
-            msg = _("Quota %s limit must be equal or greater than existing "
-                    "resources.") % key
+            msg = (_("Quota %(key)s limit must be equal or greater than "
+                     "existing resources. Current usage is %(usage)s "
+                     "and the requested limit is %(limit)s.")
+                   % {'key': key,
+                      'usage': used,
+                      'limit': value})
             raise webob.exc.HTTPBadRequest(explanation=msg)
 
     def _get_quotas(self, context, id, usages=False):
         values = QUOTAS.get_project_quotas(context, id, usages=usages)
+        group_values = GROUP_QUOTAS.get_project_quotas(context, id,
+                                                       usages=usages)
+        values.update(group_values)
 
         if usages:
             return values
@@ -98,6 +89,10 @@ class QuotaSetsController(wsgi.Controller):
         :param parent_id: The parent id of the project in which the user
                           want to perform an update or delete operation.
         """
+        if context_project.is_admin_project:
+            # The calling project has admin privileges and should be able
+            # to operate on all quotas.
+            return
         if context_project.parent_id and parent_id != context_project.id:
             msg = _("Update and delete quota operations can only be made "
                     "by an admin of immediate parent or by the CLOUD admin.")
@@ -118,15 +113,20 @@ class QuotaSetsController(wsgi.Controller):
     def _authorize_show(self, context_project, target_project):
         """Checks if show is allowed in the current hierarchy.
 
-        With hierarchical projects, are allowed to perform quota show operation
-        users with admin role in, at least, one of the following projects: the
-        current project; the immediate parent project; or the root project.
+        With hierarchical projects, users are allowed to perform a quota show
+        operation if they have the cloud admin role or if they belong to at
+        least one of the following projects: the target project, its immediate
+        parent project, or the root project of its hierarchy.
 
         :param context_project: The project in which the user
                                 is scoped to.
         :param target_project: The project in which the user wants
                                to perform a show operation.
         """
+        if context_project.is_admin_project:
+            # The calling project has admin privileges and should be able
+            # to view all quotas.
+            return
         if target_project.parent_id:
             if target_project.id != context_project.id:
                 if not self._is_descendant(target_project.id,
@@ -155,7 +155,6 @@ class QuotaSetsController(wsgi.Controller):
                     return True
         return False
 
-    @wsgi.serializers(xml=QuotaTemplate)
     def show(self, req, id):
         """Show quota for a particular tenant
 
@@ -168,12 +167,13 @@ class QuotaSetsController(wsgi.Controller):
         :param id: target project id that needs to be shown
         """
         context = req.environ['cinder.context']
-        authorize_show(context)
         params = req.params
         target_project_id = id
+        context.authorize(policy.SHOW_POLICY,
+                          target={'project_id': target_project_id})
 
         if not hasattr(params, '__call__') and 'usage' in params:
-            usage = strutils.bool_from_string(params['usage'])
+            usage = utils.get_bool_param('usage', params)
         else:
             usage = False
 
@@ -184,20 +184,15 @@ class QuotaSetsController(wsgi.Controller):
             target_project = quota_utils.get_project_hierarchy(
                 context, target_project_id)
             context_project = quota_utils.get_project_hierarchy(
-                context, context.project_id, subtree_as_ids=True)
+                context, context.project_id, subtree_as_ids=True,
+                is_admin_project=context.is_admin)
 
             self._authorize_show(context_project, target_project)
 
-        try:
-            sqlalchemy_api.authorize_project_context(context,
-                                                     target_project_id)
-        except exception.NotAuthorized:
-            raise webob.exc.HTTPForbidden()
-
         quotas = self._get_quotas(context, target_project_id, usage)
         return self._format_quota_set(target_project_id, quotas)
 
-    @wsgi.serializers(xml=QuotaTemplate)
+    @validation.schema(quotas.update_quota)
     def update(self, req, id, body):
         """Update Quota for a particular tenant
 
@@ -207,39 +202,16 @@ class QuotaSetsController(wsgi.Controller):
 
         :param req: request
         :param id: target project id that needs to be updated
-        :param body: key, value pair that that will be
-                     applied to the resources if the update
-                     succeeds
+        :param body: key, value pair that will be applied to
+                     the resources if the update succeeds
         """
         context = req.environ['cinder.context']
-        authorize_update(context)
+        target_project_id = id
+        context.authorize(policy.UPDATE_POLICY,
+                          target={'project_id': target_project_id})
         self.validate_string_length(id, 'quota_set_name',
                                     min_length=1, max_length=255)
 
-        self.assert_valid_body(body, 'quota_set')
-
-        # Get the optional argument 'skip_validation' from body,
-        # if skip_validation is False, then validate existing resource.
-        skip_flag = body.get('skip_validation', True)
-        if not utils.is_valid_boolstr(skip_flag):
-            msg = _("Invalid value '%s' for skip_validation.") % skip_flag
-            raise exception.InvalidParameterValue(err=msg)
-        skip_flag = strutils.bool_from_string(skip_flag)
-
-        target_project_id = id
-        bad_keys = []
-
-        # NOTE(ankit): Pass #1 - In this loop for body['quota_set'].items(),
-        # we figure out if we have any bad keys.
-        for key, value in body['quota_set'].items():
-            if (key not in QUOTAS and key not in NON_QUOTA_KEYS):
-                bad_keys.append(key)
-                continue
-
-        if len(bad_keys) > 0:
-            msg = _("Bad key(s) in quota set: %s") % ",".join(bad_keys)
-            raise webob.exc.HTTPBadRequest(explanation=msg)
-
         # Saving off this value since we need to use it multiple times
         use_nested_quotas = QUOTAS.using_nested_quotas()
         if use_nested_quotas:
@@ -253,45 +225,44 @@ class QuotaSetsController(wsgi.Controller):
                 # Get the children of the project which the token is scoped to
                 # in order to know if the target_project is in its hierarchy.
                 context_project = quota_utils.get_project_hierarchy(
-                    context, context.project_id, subtree_as_ids=True)
+                    context, context.project_id, subtree_as_ids=True,
+                    is_admin_project=context.is_admin)
                 self._authorize_update_or_delete(context_project,
                                                  target_project.id,
                                                  parent_id)
 
-        # NOTE(ankit): Pass #2 - In this loop for body['quota_set'].keys(),
+        # NOTE(ankit): Pass #1 - In this loop for body['quota_set'].keys(),
         # we validate the quota limits to ensure that we can bail out if
         # any of the items in the set is bad. Meanwhile we validate value
         # to ensure that the value can't be lower than number of existing
         # resources.
         quota_values = QUOTAS.get_project_quotas(context, target_project_id,
                                                  defaults=False)
+        group_quota_values = GROUP_QUOTAS.get_project_quotas(context,
+                                                             target_project_id,
+                                                             defaults=False)
+        quota_values.update(group_quota_values)
         valid_quotas = {}
         reservations = []
         for key in body['quota_set'].keys():
             if key in NON_QUOTA_KEYS:
                 continue
-
-            value = utils.validate_integer(
-                body['quota_set'][key], key, min_value=-1,
-                max_value=db.MAX_INT)
-
-            # Can't skip the validation of nested quotas since it could mess up
-            # hierarchy if parent limit is less than childrens' current usage
-            if not skip_flag or use_nested_quotas:
-                self._validate_existing_resource(key, value, quota_values)
+            self._validate_existing_resource(key, body['quota_set'][key],
+                                             quota_values)
 
             if use_nested_quotas:
                 try:
                     reservations += self._update_nested_quota_allocated(
-                        context, target_project, quota_values, key, value)
+                        context, target_project, quota_values, key,
+                        body['quota_set'][key])
                 except exception.OverQuota as e:
                     if reservations:
                         db.reservation_rollback(context, reservations)
-                    raise webob.exc.HTTPBadRequest(explanation=e.message)
+                    raise webob.exc.HTTPBadRequest(explanation=e.msg)
 
-            valid_quotas[key] = value
+            valid_quotas[key] = body['quota_set'][key]
 
-        # NOTE(ankit): Pass #3 - At this point we know that all the keys and
+        # NOTE(ankit): Pass #2 - At this point we know that all the keys and
         # values are valid and we can iterate and update them all in one shot
         # without having to worry about rolling back etc as we have done
         # the validation up front in the 2 loops above.
@@ -332,20 +303,21 @@ class QuotaSetsController(wsgi.Controller):
         res_change = new_quota_from_target_proj - orig_quota_from_target_proj
         if res_change != 0:
             deltas = {res: res_change}
+            resources = QUOTAS.resources
+            resources.update(GROUP_QUOTAS.resources)
             reservations += quota_utils.update_alloc_to_next_hard_limit(
-                ctxt, QUOTAS.resources, deltas, res, None, target_project.id)
+                ctxt, resources, deltas, res, None, target_project.id)
 
         return reservations
 
-    @wsgi.serializers(xml=QuotaTemplate)
     def defaults(self, req, id):
         context = req.environ['cinder.context']
-        authorize_show(context)
+        context.authorize(policy.SHOW_POLICY, target={'project_id': id})
+        defaults = QUOTAS.get_defaults(context, project_id=id)
+        group_defaults = GROUP_QUOTAS.get_defaults(context, project_id=id)
+        defaults.update(group_defaults)
+        return self._format_quota_set(id, defaults)
 
-        return self._format_quota_set(id, QUOTAS.get_defaults(
-            context, project_id=id))
-
-    @wsgi.serializers(xml=QuotaTemplate)
     def delete(self, req, id):
         """Delete Quota for a particular tenant.
 
@@ -357,15 +329,12 @@ class QuotaSetsController(wsgi.Controller):
         :param id: target project id that needs to be deleted
         """
         context = req.environ['cinder.context']
-        authorize_delete(context)
+        context.authorize(policy.DELETE_POLICY, target={'project_id': id})
 
         if QUOTAS.using_nested_quotas():
             self._delete_nested_quota(context, id)
         else:
-            try:
-                db.quota_destroy_by_project(context, id)
-            except exception.AdminRequired:
-                raise webob.exc.HTTPForbidden()
+            db.quota_destroy_by_project(context, id)
 
     def _delete_nested_quota(self, ctxt, proj_id):
         # Get the parent_id of the target project to verify whether we are
@@ -374,6 +343,9 @@ class QuotaSetsController(wsgi.Controller):
         try:
             project_quotas = QUOTAS.get_project_quotas(
                 ctxt, proj_id, usages=True, defaults=False)
+            project_group_quotas = GROUP_QUOTAS.get_project_quotas(
+                ctxt, proj_id, usages=True, defaults=False)
+            project_quotas.update(project_group_quotas)
         except exception.NotAuthorized:
             raise webob.exc.HTTPForbidden()
 
@@ -390,12 +362,13 @@ class QuotaSetsController(wsgi.Controller):
                                              parent_id)
 
         defaults = QUOTAS.get_defaults(ctxt, proj_id)
+        defaults.update(GROUP_QUOTAS.get_defaults(ctxt, proj_id))
         # If the project which is being deleted has allocated part of its
         # quota to its subprojects, then subprojects' quotas should be
         # deleted first.
         for res, value in project_quotas.items():
             if 'allocated' in project_quotas[res].keys():
-                if project_quotas[res]['allocated'] != 0:
+                if project_quotas[res]['allocated'] > 0:
                     msg = _("About to delete child projects having "
                             "non-zero quota. This should not be performed")
                     raise webob.exc.HTTPBadRequest(explanation=msg)
@@ -403,10 +376,7 @@ class QuotaSetsController(wsgi.Controller):
             self._validate_existing_resource(
                 res, defaults[res], project_quotas)
 
-        try:
-            db.quota_destroy_by_project(ctxt, target_project.id)
-        except exception.AdminRequired:
-            raise webob.exc.HTTPForbidden()
+        db.quota_destroy_by_project(ctxt, target_project.id)
 
         for res, limit in project_quotas.items():
             # Update child limit to 0 so the parent hierarchy gets it's
@@ -422,11 +392,22 @@ class QuotaSetsController(wsgi.Controller):
         no child quota would be larger than it's parent).
         """
         ctxt = req.environ['cinder.context']
+        ctxt.authorize(policy.VALIDATE_NESTED_QUOTA_POLICY)
         params = req.params
         try:
+            resources = QUOTAS.resources
+            resources.update(GROUP_QUOTAS.resources)
+            allocated = params.get('fix_allocated_quotas', 'False')
+            try:
+                fix_allocated = strutils.bool_from_string(allocated,
+                                                          strict=True)
+            except ValueError:
+                msg = _("Invalid param 'fix_allocated_quotas':%s") % allocated
+                raise webob.exc.HTTPBadRequest(explanation=msg)
+
             quota_utils.validate_setup_for_nested_quota_use(
-                ctxt, QUOTAS.resources, quota.NestedDbQuotaDriver(),
-                fix_allocated_quotas=params.get('fix_allocated_quotas'))
+                ctxt, resources, quota.NestedDbQuotaDriver(),
+                fix_allocated_quotas=fix_allocated)
         except exception.InvalidNestedQuotaSetup as e:
             raise webob.exc.HTTPBadRequest(explanation=e.msg)
 
@@ -436,7 +417,6 @@ class Quotas(extensions.ExtensionDescriptor):
 
     name = "Quotas"
     alias = "os-quota-sets"
-    namespace = "http://docs.openstack.org/volume/ext/quotas-sets/api/v1.1"
     updated = "2011-08-08T00:00:00+00:00"
 
     def get_resources(self):
