@@ -36,7 +36,7 @@ class CinderObjectVersionsHistory(dict):
     """Helper class that maintains objects version history.
 
     Current state of object versions is aggregated in a single version number
-    that explicitily identifies a set of object versions. That way a service
+    that explicitly identifies a set of object versions. That way a service
     is able to report what objects it supports using a single string and all
     the newer services will know exactly what that mean for a single object.
     """
@@ -116,6 +116,35 @@ OBJ_VERSIONS.add('1.10', {'Group': '1.0', 'GroupList': '1.0', 'Volume': '1.5',
                           'RequestSpec': '1.1', 'VolumeProperties': '1.1'})
 OBJ_VERSIONS.add('1.11', {'GroupSnapshot': '1.0', 'GroupSnapshotList': '1.0',
                           'Group': '1.1'})
+OBJ_VERSIONS.add('1.12', {'VolumeType': '1.3'})
+OBJ_VERSIONS.add('1.13', {'CleanupRequest': '1.0'})
+OBJ_VERSIONS.add('1.14', {'VolumeAttachmentList': '1.1'})
+OBJ_VERSIONS.add('1.15', {'Volume': '1.6', 'Snapshot': '1.2'})
+OBJ_VERSIONS.add('1.16', {'BackupDeviceInfo': '1.0'})
+OBJ_VERSIONS.add('1.17', {'VolumeAttachment': '1.1'})
+OBJ_VERSIONS.add('1.18', {'Snapshot': '1.3'})
+OBJ_VERSIONS.add('1.19', {'ConsistencyGroup': '1.4', 'CGSnapshot': '1.1'})
+OBJ_VERSIONS.add('1.20', {'Cluster': '1.1'})
+OBJ_VERSIONS.add('1.21', {'ManageableSnapshot': '1.0',
+                          'ManageableVolume': '1.0',
+                          'ManageableVolumeList': '1.0',
+                          'ManageableSnapshotList': '1.0'})
+OBJ_VERSIONS.add('1.22', {'Snapshot': '1.4'})
+OBJ_VERSIONS.add('1.23', {'VolumeAttachment': '1.2'})
+OBJ_VERSIONS.add('1.24', {'LogLevel': '1.0', 'LogLevelList': '1.0'})
+OBJ_VERSIONS.add('1.25', {'Group': '1.2'})
+OBJ_VERSIONS.add('1.26', {'Snapshot': '1.5'})
+OBJ_VERSIONS.add('1.27', {'Backup': '1.5', 'BackupImport': '1.5'})
+OBJ_VERSIONS.add('1.28', {'Service': '1.5'})
+OBJ_VERSIONS.add('1.29', {'Service': '1.6'})
+OBJ_VERSIONS.add('1.30', {'RequestSpec': '1.2'})
+OBJ_VERSIONS.add('1.31', {'Volume': '1.7'})
+OBJ_VERSIONS.add('1.32', {'RequestSpec': '1.3'})
+OBJ_VERSIONS.add('1.33', {'Volume': '1.8'})
+OBJ_VERSIONS.add('1.34', {'VolumeAttachment': '1.3'})
+OBJ_VERSIONS.add('1.35', {'Backup': '1.6', 'BackupImport': '1.6'})
+OBJ_VERSIONS.add('1.36', {'RequestSpec': '1.4'})
+OBJ_VERSIONS.add('1.37', {'RequestSpec': '1.5'})
 
 
 class CinderObjectRegistry(base.VersionedObjectRegistry):
@@ -271,6 +300,34 @@ class CinderPersistentObject(object):
         finally:
             self._context = original_context
 
+    @contextlib.contextmanager
+    def as_read_deleted(self, mode='yes'):
+        """Context manager to make OVO with modified read deleted context.
+
+        This temporarily modifies the context embedded in an object to
+        have a different `read_deleted` parameter.
+
+        Parameter mode accepts most of the same parameters as our `model_query`
+        DB method.  We support 'yes', 'no', and 'only'.
+
+        usage:
+
+           with obj.as_read_deleted():
+               obj.refresh()
+           if obj.status = 'deleted':
+               ...
+        """
+        if self._context is None:
+            raise exception.OrphanedObjectError(method='as_read_deleted',
+                                                objtype=self.obj_name())
+
+        original_mode = self._context.read_deleted
+        self._context.read_deleted = mode
+        try:
+            yield
+        finally:
+            self._context.read_deleted = original_mode
+
     @classmethod
     def _get_expected_attrs(cls, context, *args, **kwargs):
         return None
@@ -285,7 +342,6 @@ class CinderPersistentObject(object):
             raise NotImplementedError(msg)
 
         orm_obj = db.get_by_id(context, cls.model, id, *args, **kwargs)
-        expected_attrs = cls._get_expected_attrs(context)
         # We pass parameters because fields to expect may depend on them
         expected_attrs = cls._get_expected_attrs(context, *args, **kwargs)
         kargs = {}
@@ -293,6 +349,12 @@ class CinderPersistentObject(object):
             kargs = {'expected_attrs': expected_attrs}
         return cls._from_db_object(context, cls(context), orm_obj, **kargs)
 
+    def update_single_status_where(self, new_status,
+                                   expected_status, filters=()):
+        values = {'status': new_status}
+        expected_status = {'status': expected_status}
+        return self.conditional_update(values, expected_status, filters)
+
     def conditional_update(self, values, expected_values=None, filters=(),
                            save_all=False, session=None, reflect_changes=True,
                            order=None):
@@ -453,6 +515,17 @@ class ClusteredObject(object):
     def service_topic_queue(self):
         return self.cluster_name or self.host
 
+    @property
+    def is_clustered(self):
+        return bool(self.cluster_name)
+
+    def assert_not_frozen(self):
+        ctxt = self._context.elevated()
+        if db.is_backend_frozen(ctxt, self.host, self.cluster_name):
+            msg = _('Modification operations are not allowed on frozen '
+                    'storage backends.')
+            raise exception.InvalidInput(reason=msg)
+
 
 class CinderObjectSerializer(base.VersionedObjectSerializer):
     OBJ_BASE_CLASS = CinderObject
