@@ -18,6 +18,7 @@ import uuid
 import mock
 from oslo_config import cfg
 from oslo_serialization import jsonutils
+from six.moves import http_client
 import webob
 
 from cinder.api import extensions
@@ -28,7 +29,7 @@ from cinder import exception
 from cinder import objects
 from cinder import test
 from cinder.tests.unit.api import fakes
-from cinder.tests.unit.api.v2 import stubs
+from cinder.tests.unit.api.v2 import fakes as v2_fakes
 from cinder.tests.unit import fake_constants as fake
 from cinder.tests.unit import fake_volume
 from cinder import volume
@@ -39,43 +40,30 @@ CONF = cfg.CONF
 
 
 def return_create_volume_metadata_max(context, volume_id, metadata, delete):
-    return stub_max_volume_metadata()
+    return fake_max_volume_metadata()
 
 
 def return_create_volume_metadata(context, volume_id, metadata,
                                   delete, meta_type):
-    return stub_volume_metadata()
+    return fake_volume_metadata()
 
 
 def return_new_volume_metadata(context, volume_id, metadata,
                                delete, meta_type):
-    return stub_new_volume_metadata()
+    return fake_new_volume_metadata()
 
 
 def return_create_volume_metadata_insensitive(context, snapshot_id,
                                               metadata, delete,
                                               meta_type):
-    return stub_volume_metadata_insensitive()
+    return fake_volume_metadata_insensitive()
 
 
 def return_volume_metadata(context, volume_id):
-    return stub_volume_metadata()
+    return fake_volume_metadata()
 
 
-def return_empty_volume_metadata(context, volume_id):
-    return {}
-
-
-def return_empty_container_metadata(context, volume_id, metadata,
-                                    delete, meta_type):
-    return {}
-
-
-def delete_volume_metadata(context, volume_id, key, meta_type):
-    pass
-
-
-def stub_volume_metadata():
+def fake_volume_metadata():
     metadata = {
         "key1": "value1",
         "key2": "value2",
@@ -84,7 +72,7 @@ def stub_volume_metadata():
     return metadata
 
 
-def stub_new_volume_metadata():
+def fake_new_volume_metadata():
     metadata = {
         'key10': 'value10',
         'key99': 'value99',
@@ -93,7 +81,7 @@ def stub_new_volume_metadata():
     return metadata
 
 
-def stub_volume_metadata_insensitive():
+def fake_volume_metadata_insensitive():
     metadata = {
         "key1": "value1",
         "key2": "value2",
@@ -103,7 +91,7 @@ def stub_volume_metadata_insensitive():
     return metadata
 
 
-def stub_max_volume_metadata():
+def fake_max_volume_metadata():
     metadata = {"metadata": {}}
     for num in range(CONF.quota_metadata_items):
         metadata['metadata']['key%i' % num] = "blah"
@@ -120,24 +108,19 @@ def return_volume_nonexistent(*args, **kwargs):
     raise exception.VolumeNotFound('bogus test message')
 
 
-def fake_update_volume_metadata(self, context, volume, diff):
-    pass
-
-
-class volumeMetaDataTest(test.TestCase):
+class VolumeMetaDataTest(test.TestCase):
 
     def setUp(self):
-        super(volumeMetaDataTest, self).setUp()
+        super(VolumeMetaDataTest, self).setUp()
         self.volume_api = volume_api.API()
-        self.stubs.Set(volume.api.API, 'get', get_volume)
-        self.stubs.Set(db, 'volume_metadata_get',
-                       return_volume_metadata)
-        self.patch(
-            'cinder.db.service_get_all', autospec=True,
-            return_value=stubs.stub_service_get_all_by_topic(None, None))
-
-        self.stubs.Set(self.volume_api, 'update_volume_metadata',
-                       fake_update_volume_metadata)
+        self.mock_object(volume.api.API, 'get', get_volume)
+        self.mock_object(db, 'volume_metadata_get',
+                         return_volume_metadata)
+        self.mock_object(db, 'service_get_all',
+                         return_value=v2_fakes.fake_service_get_all_by_topic(
+                             None, None),
+                         autospec=True)
+        self.mock_object(self.volume_api, 'update_volume_metadata')
 
         self.ext_mgr = extensions.ExtensionManager()
         self.ext_mgr.extensions = {}
@@ -154,7 +137,7 @@ class volumeMetaDataTest(test.TestCase):
                "metadata": {}}
         body = {"volume": vol}
         req = fakes.HTTPRequest.blank('/v2/volumes')
-        self.volume_controller.create(req, body)
+        self.volume_controller.create(req, body=body)
 
     def test_index(self):
         req = fakes.HTTPRequest.blank(self.url)
@@ -170,15 +153,15 @@ class volumeMetaDataTest(test.TestCase):
         self.assertEqual(expected, res_dict)
 
     def test_index_nonexistent_volume(self):
-        self.stubs.Set(db, 'volume_metadata_get',
-                       return_volume_nonexistent)
+        self.mock_object(db, 'volume_metadata_get',
+                         return_volume_nonexistent)
         req = fakes.HTTPRequest.blank(self.url)
         self.assertRaises(exception.VolumeNotFound,
                           self.controller.index, req, self.url)
 
     def test_index_no_data(self):
-        self.stubs.Set(db, 'volume_metadata_get',
-                       return_empty_volume_metadata)
+        self.mock_object(db, 'volume_metadata_get',
+                         return_value={})
         req = fakes.HTTPRequest.blank(self.url)
         res_dict = self.controller.index(req, self.req_id)
         expected = {'metadata': {}}
@@ -191,15 +174,15 @@ class volumeMetaDataTest(test.TestCase):
         self.assertEqual(expected, res_dict)
 
     def test_show_nonexistent_volume(self):
-        self.stubs.Set(db, 'volume_metadata_get',
-                       return_volume_nonexistent)
+        self.mock_object(db, 'volume_metadata_get',
+                         return_volume_nonexistent)
         req = fakes.HTTPRequest.blank(self.url + '/key2')
         self.assertRaises(exception.VolumeNotFound,
                           self.controller.show, req, self.req_id, 'key2')
 
     def test_show_meta_not_found(self):
-        self.stubs.Set(db, 'volume_metadata_get',
-                       return_empty_volume_metadata)
+        self.mock_object(db, 'volume_metadata_get',
+                         return_value={})
         req = fakes.HTTPRequest.blank(self.url + '/key6')
         self.assertRaises(exception.VolumeMetadataNotFound,
                           self.controller.show, req, self.req_id, 'key6')
@@ -210,7 +193,6 @@ class volumeMetaDataTest(test.TestCase):
         fake_volume = objects.Volume(id=self.req_id, status='available')
         fake_context = mock.Mock()
         metadata_get.side_effect = return_volume_metadata
-        metadata_delete.side_effect = delete_volume_metadata
         req = fakes.HTTPRequest.blank(self.url + '/key2')
         req.method = 'DELETE'
         req.environ['cinder.context'] = fake_context
@@ -219,7 +201,7 @@ class volumeMetaDataTest(test.TestCase):
                                'get') as get_volume:
             get_volume.return_value = fake_volume
             res = self.controller.delete(req, self.req_id, 'key2')
-            self.assertEqual(200, res.status_int)
+            self.assertEqual(http_client.OK, res.status_int)
             get_volume.assert_called_once_with(fake_context, self.req_id)
 
     @mock.patch.object(db, 'volume_metadata_delete')
@@ -228,7 +210,6 @@ class volumeMetaDataTest(test.TestCase):
         fake_volume = objects.Volume(id=self.req_id, status='maintenance')
         fake_context = mock.Mock()
         metadata_get.side_effect = return_volume_metadata
-        metadata_delete.side_effect = delete_volume_metadata
         req = fakes.HTTPRequest.blank(self.url + '/key2')
         req.method = 'DELETE'
         req.environ['cinder.context'] = fake_context
@@ -261,8 +242,8 @@ class volumeMetaDataTest(test.TestCase):
             get_volume.assert_called_once_with(fake_context, self.req_id)
 
     def test_delete_meta_not_found(self):
-        self.stubs.Set(db, 'volume_metadata_get',
-                       return_empty_volume_metadata)
+        self.mock_object(db, 'volume_metadata_get',
+                         return_value={})
         req = fakes.HTTPRequest.blank(self.url + '/key6')
         req.method = 'DELETE'
         self.assertRaises(exception.VolumeMetadataNotFound,
@@ -273,7 +254,7 @@ class volumeMetaDataTest(test.TestCase):
     def test_create(self, metadata_get, metadata_update):
         fake_volume = {'id': self.req_id, 'status': 'available'}
         fake_context = mock.Mock()
-        metadata_get.side_effect = return_empty_volume_metadata
+        metadata_get.return_value = {}
         metadata_update.side_effect = return_create_volume_metadata
         req = fakes.HTTPRequest.blank('/v2/volume_metadata')
         req.method = 'POST'
@@ -287,7 +268,7 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            res_dict = self.controller.create(req, self.req_id, body)
+            res_dict = self.controller.create(req, self.req_id, body=body)
             self.assertEqual(body, res_dict)
 
     @mock.patch.object(db, 'volume_metadata_update')
@@ -295,7 +276,7 @@ class volumeMetaDataTest(test.TestCase):
     def test_create_volume_maintenance(self, metadata_get, metadata_update):
         fake_volume = {'id': self.req_id, 'status': 'maintenance'}
         fake_context = mock.Mock()
-        metadata_get.side_effect = return_empty_volume_metadata
+        metadata_get.return_value = {}
         metadata_update.side_effect = return_create_volume_metadata
         req = fakes.HTTPRequest.blank('/v2/volume_metadata')
         req.method = 'POST'
@@ -311,7 +292,7 @@ class volumeMetaDataTest(test.TestCase):
             get_volume.return_value = fake_volume
             self.assertRaises(exception.InvalidVolume,
                               self.controller.create,
-                              req, self.req_id, body)
+                              req, self.req_id, body=body)
 
     @mock.patch.object(db, 'volume_metadata_update')
     @mock.patch.object(db, 'volume_metadata_get')
@@ -321,7 +302,7 @@ class volumeMetaDataTest(test.TestCase):
         # which server added
         fake_volume = {'id': self.req_id, 'status': 'available'}
         fake_context = mock.Mock()
-        metadata_get.side_effect = return_empty_volume_metadata
+        metadata_get.return_value = {}
         metadata_update.side_effect = return_create_volume_metadata_insensitive
 
         req = fakes.HTTPRequest.blank('/v2/volume_metadata')
@@ -343,50 +324,61 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            res_dict = self.controller.create(req, self.req_id, body)
+            res_dict = self.controller.create(req, self.req_id, body=body)
             self.assertEqual(expected, res_dict)
 
     def test_create_empty_body(self):
-        self.stubs.Set(db, 'volume_metadata_update',
-                       return_create_volume_metadata)
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
         req = fakes.HTTPRequest.blank(self.url)
         req.method = 'POST'
         req.headers["content-type"] = "application/json"
 
-        self.assertRaises(webob.exc.HTTPBadRequest,
-                          self.controller.create, req, self.req_id, None)
+        self.assertRaises(exception.ValidationError,
+                          self.controller.create, req, self.req_id, body=None)
+
+    def test_create_metadata_keys_value_none(self):
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
+        req = fakes.HTTPRequest.blank(self.url)
+        req.method = 'POST'
+        req.headers["content-type"] = "application/json"
+        body = {"meta": {"key": None}}
+        self.assertRaises(exception.ValidationError,
+                          self.controller.create, req, self.req_id, body=body)
 
     def test_create_item_empty_key(self):
-        self.stubs.Set(db, 'volume_metadata_update',
-                       return_create_volume_metadata)
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
         req = fakes.HTTPRequest.blank(self.url + '/key1')
         req.method = 'PUT'
         body = {"meta": {"": "value1"}}
         req.body = jsonutils.dump_as_bytes(body)
         req.headers["content-type"] = "application/json"
 
-        self.assertRaises(webob.exc.HTTPBadRequest,
-                          self.controller.create, req, self.req_id, body)
+        self.assertRaises(exception.ValidationError,
+                          self.controller.create, req, self.req_id, body=body)
 
     def test_create_item_key_too_long(self):
-        self.stubs.Set(db, 'volume_metadata_update',
-                       return_create_volume_metadata)
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
         req = fakes.HTTPRequest.blank(self.url + '/key1')
         req.method = 'PUT'
         body = {"meta": {("a" * 260): "value1"}}
         req.body = jsonutils.dump_as_bytes(body)
         req.headers["content-type"] = "application/json"
 
-        self.assertRaises(webob.exc.HTTPBadRequest,
+        self.assertRaises(exception.ValidationError,
                           self.controller.create,
-                          req, self.req_id, body)
+                          req, self.req_id, body=body)
 
     def test_create_nonexistent_volume(self):
-        self.stubs.Set(volume.api.API, 'get', return_volume_nonexistent)
-        self.stubs.Set(db, 'volume_metadata_get',
-                       return_volume_metadata)
-        self.stubs.Set(db, 'volume_metadata_update',
-                       return_create_volume_metadata)
+        self.mock_object(volume.api.API, 'get',
+                         return_volume_nonexistent)
+        self.mock_object(db, 'volume_metadata_get',
+                         return_volume_metadata)
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
 
         req = fakes.HTTPRequest.blank('/v2/volume_metadata')
         req.method = 'POST'
@@ -394,7 +386,7 @@ class volumeMetaDataTest(test.TestCase):
         body = {"metadata": {"key9": "value9"}}
         req.body = jsonutils.dump_as_bytes(body)
         self.assertRaises(exception.VolumeNotFound,
-                          self.controller.create, req, self.req_id, body)
+                          self.controller.create, req, self.req_id, body=body)
 
     @mock.patch.object(db, 'volume_metadata_update')
     def test_update_all(self, metadata_update):
@@ -417,7 +409,8 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            res_dict = self.controller.update_all(req, self.req_id, expected)
+            res_dict = self.controller.update_all(req, self.req_id,
+                                                  body=expected)
             self.assertEqual(expected, res_dict)
             get_volume.assert_called_once_with(fake_context, self.req_id)
 
@@ -444,7 +437,7 @@ class volumeMetaDataTest(test.TestCase):
             get_volume.return_value = fake_volume
             self.assertRaises(exception.InvalidVolume,
                               self.controller.update_all, req,
-                              self.req_id, expected)
+                              self.req_id, body=expected)
             self.assertFalse(metadata_update.called)
             get_volume.assert_called_once_with(fake_context, self.req_id)
 
@@ -481,7 +474,7 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            res_dict = self.controller.update_all(req, self.req_id, body)
+            res_dict = self.controller.update_all(req, self.req_id, body=body)
             self.assertEqual(expected, res_dict)
             get_volume.assert_called_once_with(fake_context, self.req_id)
 
@@ -489,7 +482,7 @@ class volumeMetaDataTest(test.TestCase):
     def test_update_all_empty_container(self, metadata_update):
         fake_volume = {'id': self.req_id, 'status': 'available'}
         fake_context = mock.Mock()
-        metadata_update.side_effect = return_empty_container_metadata
+        metadata_update.return_value = {}
         req = fakes.HTTPRequest.blank(self.url)
         req.method = 'PUT'
         req.content_type = "application/json"
@@ -500,38 +493,39 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            res_dict = self.controller.update_all(req, self.req_id, expected)
+            res_dict = self.controller.update_all(req, self.req_id,
+                                                  body=expected)
             self.assertEqual(expected, res_dict)
             get_volume.assert_called_once_with(fake_context, self.req_id)
 
     def test_update_all_malformed_container(self):
-        self.stubs.Set(db, 'volume_metadata_update',
-                       return_create_volume_metadata)
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
         req = fakes.HTTPRequest.blank(self.url)
         req.method = 'PUT'
         req.content_type = "application/json"
         expected = {'meta': {}}
         req.body = jsonutils.dump_as_bytes(expected)
 
-        self.assertRaises(webob.exc.HTTPBadRequest,
+        self.assertRaises(exception.ValidationError,
                           self.controller.update_all, req, self.req_id,
-                          expected)
+                          body=expected)
 
     def test_update_all_malformed_data(self):
-        self.stubs.Set(db, 'volume_metadata_update',
-                       return_create_volume_metadata)
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
         req = fakes.HTTPRequest.blank(self.url)
         req.method = 'PUT'
         req.content_type = "application/json"
         expected = {'metadata': ['asdf']}
         req.body = jsonutils.dump_as_bytes(expected)
 
-        self.assertRaises(webob.exc.HTTPBadRequest,
+        self.assertRaises(exception.ValidationError,
                           self.controller.update_all, req, self.req_id,
-                          expected)
+                          body=expected)
 
     def test_update_all_nonexistent_volume(self):
-        self.stubs.Set(db, 'volume_get', return_volume_nonexistent)
+        self.mock_object(db, 'volume_get', return_volume_nonexistent)
         req = fakes.HTTPRequest.blank(self.url)
         req.method = 'PUT'
         req.content_type = "application/json"
@@ -539,7 +533,7 @@ class volumeMetaDataTest(test.TestCase):
         req.body = jsonutils.dump_as_bytes(body)
 
         self.assertRaises(exception.VolumeNotFound,
-                          self.controller.update_all, req, '100', body)
+                          self.controller.update_all, req, '100', body=body)
 
     @mock.patch.object(db, 'volume_metadata_update')
     def test_update_item(self, metadata_update):
@@ -556,11 +550,25 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            res_dict = self.controller.update(req, self.req_id, 'key1', body)
+            res_dict = self.controller.update(req, self.req_id, 'key1',
+                                              body=body)
             expected = {'meta': {'key1': 'value1'}}
             self.assertEqual(expected, res_dict)
             get_volume.assert_called_once_with(fake_context, self.req_id)
 
+    def test_update_metadata_item_keys_value_none(self):
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
+        req = fakes.HTTPRequest.blank(self.url + '/key1')
+        req.method = 'PUT'
+        body = {"meta": {"a": None}}
+        req.body = jsonutils.dump_as_bytes(body)
+        req.headers["content-type"] = "application/json"
+
+        self.assertRaises(exception.ValidationError,
+                          self.controller.update,
+                          req, self.req_id, 'key1', body=body)
+
     @mock.patch.object(db, 'volume_metadata_update')
     def test_update_item_volume_maintenance(self, metadata_update):
         fake_volume = {'id': self.req_id, 'status': 'maintenance'}
@@ -578,13 +586,13 @@ class volumeMetaDataTest(test.TestCase):
             get_volume.return_value = fake_volume
             self.assertRaises(exception.InvalidVolume,
                               self.controller.update, req,
-                              self.req_id, 'key1', body)
+                              self.req_id, 'key1', body=body)
             self.assertFalse(metadata_update.called)
             get_volume.assert_called_once_with(fake_context, self.req_id)
 
     def test_update_item_nonexistent_volume(self):
-        self.stubs.Set(db, 'volume_get',
-                       return_volume_nonexistent)
+        self.mock_object(db, 'volume_get',
+                         return_volume_nonexistent)
         req = fakes.HTTPRequest.blank(
             '/v2/%s/volumes/asdf/metadata/key1' % fake.PROJECT_ID)
         req.method = 'PUT'
@@ -594,18 +602,18 @@ class volumeMetaDataTest(test.TestCase):
 
         self.assertRaises(exception.VolumeNotFound,
                           self.controller.update, req, self.req_id, 'key1',
-                          body)
+                          body=body)
 
     def test_update_item_empty_body(self):
-        self.stubs.Set(db, 'volume_metadata_update',
-                       return_create_volume_metadata)
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
         req = fakes.HTTPRequest.blank(self.url + '/key1')
         req.method = 'PUT'
         req.headers["content-type"] = "application/json"
 
-        self.assertRaises(webob.exc.HTTPBadRequest,
+        self.assertRaises(exception.ValidationError,
                           self.controller.update, req, self.req_id, 'key1',
-                          None)
+                          body=None)
 
     @mock.patch.object(db, 'volume_metadata_update')
     def test_update_item_empty_key(self, metadata_update):
@@ -622,11 +630,10 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            self.assertRaises(webob.exc.HTTPBadRequest,
+            self.assertRaises(exception.ValidationError,
                               self.controller.update, req, self.req_id,
-                              '', body)
+                              '', body=body)
             self.assertFalse(metadata_update.called)
-            get_volume.assert_called_once_with(fake_context, self.req_id)
 
     @mock.patch.object(db, 'volume_metadata_update')
     def test_update_item_key_too_long(self, metadata_update):
@@ -643,11 +650,10 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
+            self.assertRaises(exception.ValidationError,
                               self.controller.update,
-                              req, self.req_id, ("a" * 260), body)
+                              req, self.req_id, ("a" * 260), body=body)
             self.assertFalse(metadata_update.called)
-            get_volume.assert_called_once_with(fake_context, self.req_id)
 
     @mock.patch.object(db, 'volume_metadata_update')
     def test_update_item_value_too_long(self, metadata_update):
@@ -664,28 +670,27 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
+            self.assertRaises(exception.ValidationError,
                               self.controller.update,
-                              req, self.req_id, "key1", body)
+                              req, self.req_id, "key1", body=body)
             self.assertFalse(metadata_update.called)
-            get_volume.assert_called_once_with(fake_context, self.req_id)
 
     def test_update_item_too_many_keys(self):
-        self.stubs.Set(db, 'volume_metadata_update',
-                       return_create_volume_metadata)
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
         req = fakes.HTTPRequest.blank(self.url + '/key1')
         req.method = 'PUT'
         body = {"meta": {"key1": "value1", "key2": "value2"}}
         req.body = jsonutils.dump_as_bytes(body)
         req.headers["content-type"] = "application/json"
 
-        self.assertRaises(webob.exc.HTTPBadRequest,
+        self.assertRaises(exception.ValidationError,
                           self.controller.update, req, self.req_id, 'key1',
-                          body)
+                          body=body)
 
     def test_update_item_body_uri_mismatch(self):
-        self.stubs.Set(db, 'volume_metadata_update',
-                       return_create_volume_metadata)
+        self.mock_object(db, 'volume_metadata_update',
+                         return_create_volume_metadata)
         req = fakes.HTTPRequest.blank(self.url + '/bad')
         req.method = 'PUT'
         body = {"meta": {"key1": "value1"}}
@@ -694,7 +699,7 @@ class volumeMetaDataTest(test.TestCase):
 
         self.assertRaises(webob.exc.HTTPBadRequest,
                           self.controller.update, req, self.req_id, 'bad',
-                          body)
+                          body=body)
 
     @mock.patch.object(db, 'volume_metadata_update')
     def test_invalid_metadata_items_on_create(self, metadata_update):
@@ -713,8 +718,9 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
-                              self.controller.create, req, self.req_id, data)
+            self.assertRaises(exception.ValidationError,
+                              self.controller.create, req, self.req_id,
+                              body=data)
 
         # test for long value
         data = {"metadata": {"key": "v" * 260}}
@@ -724,8 +730,9 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
-                              self.controller.create, req, self.req_id, data)
+            self.assertRaises(exception.ValidationError,
+                              self.controller.create, req, self.req_id,
+                              body=data)
 
         # test for empty key.
         data = {"metadata": {"": "value1"}}
@@ -735,5 +742,6 @@ class volumeMetaDataTest(test.TestCase):
         with mock.patch.object(self.controller.volume_api,
                                'get') as get_volume:
             get_volume.return_value = fake_volume
-            self.assertRaises(webob.exc.HTTPBadRequest,
-                              self.controller.create, req, self.req_id, data)
+            self.assertRaises(exception.ValidationError,
+                              self.controller.create, req, self.req_id,
+                              body=data)
