@@ -15,8 +15,8 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 #
-"""
-Volume driver for HPE 3PAR Storage array.
+"""Volume driver for HPE 3PAR Storage array.
+
 This driver requires 3.1.3 or later firmware on the 3PAR array, using
 the 4.x version of the hpe3parclient.
 
@@ -109,10 +109,13 @@ class HPE3PARFCDriver(hpebasedriver.HPE3PARDriverBase):
         4.0.2 - Create one vlun in single path configuration. bug #1727176
         4.0.3 - Create FC vlun as host sees. bug #1734505
         4.0.4 - Handle force detach case. bug #1686745
+        4.0.5 - Set proper backend on subsequent operation, after group
+                failover. bug #1773069
+        4.0.6 - Set NSP for single path attachments. Bug #1809249
 
     """
 
-    VERSION = "4.0.4"
+    VERSION = "4.0.5"
 
     # The name of the CI wiki page.
     CI_WIKI_NAME = "HPE_Storage_CI"
@@ -123,7 +126,6 @@ class HPE3PARFCDriver(hpebasedriver.HPE3PARDriverBase):
         self.protocol = 'FC'
 
     @utils.trace
-    @fczm_utils.add_fc_zone
     def initialize_connection(self, volume, connector):
         """Assigns the volume to a server.
 
@@ -163,16 +165,27 @@ class HPE3PARFCDriver(hpebasedriver.HPE3PARDriverBase):
           * Create a VLUN for that HOST with the volume we want to export.
 
         """
-        common = self._login()
+        array_id = self.get_volume_replication_driver_data(volume)
+        common = self._login(array_id=array_id)
         try:
             # we have to make sure we have a host
             host = self._create_host(common, volume, connector)
-            target_wwns, init_targ_map, numPaths = \
-                self._build_initiator_target_map(common, connector)
-            if not connector.get('multipath'):
-                target_wwns = target_wwns[:1]
+            target_wwns, init_targ_map, numPaths = (
+                self._build_initiator_target_map(common, connector))
+
+            multipath = connector.get('multipath')
+            LOG.debug("multipath: %s", multipath)
+            user_target = None
+            if not multipath:
+                user_target = self._get_user_target(common)
                 initiator = connector.get('wwpns')[0]
-                init_targ_map[initiator] = init_targ_map[initiator][:1]
+                if user_target is None:
+                    target_wwns = target_wwns[:1]
+                    init_targ_map[initiator] = init_targ_map[initiator][:1]
+                else:
+                    target_wwns = [user_target]
+                    init_targ_map[initiator] = [user_target]
+
             # check if a VLUN already exists for this host
             existing_vlun = common.find_existing_vlun(volume, host)
 
@@ -200,15 +213,16 @@ class HPE3PARFCDriver(hpebasedriver.HPE3PARDriverBase):
 
             encryption_key_id = volume.get('encryption_key_id', None)
             info['data']['encrypted'] = encryption_key_id is not None
+            fczm_utils.add_fc_zone(info)
             return info
         finally:
             self._logout(common)
 
     @utils.trace
-    @fczm_utils.remove_fc_zone
     def terminate_connection(self, volume, connector, **kwargs):
         """Driver entry point to unattach a volume from an instance."""
-        common = self._login()
+        array_id = self.get_volume_replication_driver_data(volume)
+        common = self._login(array_id=array_id)
         try:
             is_force_detach = connector is None
             if is_force_detach:
@@ -247,7 +261,7 @@ class HPE3PARFCDriver(hpebasedriver.HPE3PARDriverBase):
 
                 info['data'] = {'target_wwn': target_wwns,
                                 'initiator_target_map': init_targ_map}
-
+                fczm_utils.remove_fc_zone(info)
             return info
 
         finally:
@@ -405,3 +419,25 @@ class HPE3PARFCDriver(hpebasedriver.HPE3PARDriverBase):
             self._modify_3par_fibrechan_host(common, host['name'], new_wwns)
             host = common._get_3par_host(host['name'])
         return host
+
+    def _get_user_target(self, common):
+        target_nsp = common.config.hpe3par_target_nsp
+
+        if not target_nsp:
+            return None
+
+        # Get target wwn from target nsp
+        fc_ports = common.get_active_fc_target_ports()
+
+        target_wwn = None
+        for port in fc_ports:
+            nsp = port['nsp']
+            if target_nsp == nsp:
+                target_wwn = port['portWWN']
+                break
+
+        if not target_wwn:
+            LOG.warning("Did not get wwn for target nsp: "
+                        "%(nsp)s", {'nsp': target_nsp})
+
+        return target_wwn
