@@ -30,10 +30,11 @@ import six
 
 import cinder
 from cinder import exception
-from cinder.i18n import _, _LE, _LW, _LI
+from cinder.i18n import _
 from cinder import objects
 from cinder.objects import fields
 from cinder import utils
+from cinder.volume import configuration
 from cinder.volume.drivers.san import san
 from cinder.volume import utils as vol_utils
 
@@ -45,24 +46,27 @@ MAX_K2_RETRY = 5
 K2_REP_FAILED_OVER = fields.ReplicationStatus.FAILED_OVER
 LOG = logging.getLogger(__name__)
 
-kaminario1_opts = [
-    cfg.StrOpt('kaminario_nodedup_substring',
-               default='K2-nodedup',
-               help="If volume-type name contains this substring "
-                    "nodedup volume will be created, otherwise "
-                    "dedup volume wil be created.",
-               deprecated_for_removal=True,
-               deprecated_reason="This option is deprecated in favour of "
-                                 "'kaminario:thin_prov_type' in extra-specs "
-                                 "and will be removed in the next release.")]
-kaminario2_opts = [
+kaminario_opts = [
     cfg.BoolOpt('auto_calc_max_oversubscription_ratio',
                 default=False,
                 help="K2 driver will calculate max_oversubscription_ratio "
-                     "on setting this option as True.")]
+                     "on setting this option as True."),
+    cfg.BoolOpt('unique_fqdn_network',
+                default=True,
+                help="Whether or not our private network has unique FQDN on "
+                     "each initiator or not.  For example networks with QA "
+                     "systems usually have multiple servers/VMs with the same "
+                     "FQDN.  When true this will create host entries on K2 "
+                     "using the FQDN, when false it will use the reversed "
+                     "IQN/WWNN."),
+    cfg.BoolOpt('disable_discovery',
+                default=False,
+                help="Disabling iSCSI discovery (sendtargets) for multipath "
+                     "connections on K2 driver."),
+]
 
 CONF = cfg.CONF
-CONF.register_opts(kaminario1_opts)
+CONF.register_opts(kaminario_opts, group=configuration.SHARED_CONF_GROUP)
 
 K2HTTPError = requests.exceptions.HTTPError
 K2_RETRY_ERRORS = ("MC_ERR_BUSY", "MC_ERR_BUSY_SPECIFIC",
@@ -101,28 +105,6 @@ if krest:
                 self.krestlock.release()
 
 
-def kaminario_logger(func):
-    """Return a function wrapper.
-
-    The wrapper adds log for entry and exit to the function.
-    """
-    def func_wrapper(*args, **kwargs):
-        LOG.debug('Entering %(function)s of %(class)s with arguments: '
-                  ' %(args)s, %(kwargs)s',
-                  {'class': args[0].__class__.__name__,
-                   'function': func.__name__,
-                   'args': args[1:],
-                   'kwargs': kwargs})
-        ret = func(*args, **kwargs)
-        LOG.debug('Exiting %(function)s of %(class)s '
-                  'having return value: %(ret)s',
-                  {'class': args[0].__class__.__name__,
-                   'function': func.__name__,
-                   'ret': ret})
-        return ret
-    return func_wrapper
-
-
 class Replication(object):
     def __init__(self, config, *args, **kwargs):
         self.backend_id = config.get('backend_id')
@@ -138,13 +120,13 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
     def __init__(self, *args, **kwargs):
         super(KaminarioCinderDriver, self).__init__(*args, **kwargs)
         self.configuration.append_config_values(san.san_opts)
-        self.configuration.append_config_values(kaminario2_opts)
+        self.configuration.append_config_values(kaminario_opts)
         self.replica = None
         self._protocol = None
         k2_lock_sfx = self.configuration.safe_get('san_ip')
         self.k2_lock_name = "%s-%s" % (K2_LOCK_PREFIX, k2_lock_sfx)
 
-    @kaminario_logger
+    @utils.trace
     def check_for_setup_error(self):
         if krest is None:
             msg = _("Unable to import 'krest' python module.")
@@ -167,13 +149,13 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                 ver_exist = versionutils.convert_version_to_int(ver)
                 ver_min = versionutils.convert_version_to_int(K2_MIN_VERSION)
                 if ver_exist < ver_min:
-                    msg = _("K2 rest api version should be "
+                    msg = _("K2 REST API version should be "
                             ">= %s.") % K2_MIN_VERSION
                     LOG.error(msg)
                     raise exception.KaminarioCinderDriverException(reason=msg)
 
             else:
-                msg = _("K2 rest api version search failed.")
+                msg = _("K2 REST API version search failed.")
                 LOG.error(msg)
                 raise exception.KaminarioCinderDriverException(reason=msg)
 
@@ -193,12 +175,12 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                     raise exception.InvalidInput(reason=msg)
             self.replica = Replication(replica[0])
 
-    @kaminario_logger
+    @utils.trace
     def do_setup(self, context):
         super(KaminarioCinderDriver, self).do_setup(context)
         self._check_ops()
 
-    @kaminario_logger
+    @utils.trace
     def create_volume(self, volume):
         """Volume creation in K2 needs a volume group.
 
@@ -226,14 +208,13 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             if vg_rs.total != 0:
                 LOG.debug("Deleting vg: %s for failed volume in K2.", vg_name)
                 vg_rs.hits[0].delete()
-            LOG.exception(_LE("Creation of volume %s failed."), vol_name)
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            LOG.exception("Creation of volume %s failed.", vol_name)
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
         if self._get_is_replica(volume.volume_type) and self.replica:
             self._create_volume_replica(volume, vg, vol, self.replica.rpo)
 
-    @kaminario_logger
+    @utils.trace
     def _create_volume_replica(self, volume, vg, vol, rpo):
         """Volume replica creation in K2 needs session and remote volume.
 
@@ -277,8 +258,8 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             src_ssn.state = "in_sync"
             src_ssn.save()
         except Exception as ex:
-            LOG.exception(_LE("Replication for the volume %s has "
-                              "failed."), vol.name)
+            LOG.exception("Replication for the volume %s has "
+                          "failed.", vol.name)
             self._delete_by_ref(self.client, "replication/sessions",
                                 session_name, 'session')
             self._delete_by_ref(self.target, "replication/sessions",
@@ -289,10 +270,9 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             self._delete_by_ref(self.target, "volume_groups",
                                 rvg_name, "remote vg")
             self._delete_by_ref(self.client, "volume_groups", vg.name, "vg")
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
-    @kaminario_logger
+    @utils.trace
     def _create_failover_volume_replica(self, volume, vg_name, vol_name):
         """Volume replica creation in K2 needs session and remote volume.
 
@@ -337,25 +317,24 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             tgt_ssn.state = "in_sync"
             tgt_ssn.save()
         except Exception as ex:
-            LOG.exception(_LE("Replication for the volume %s has "
-                              "failed."), rvol_name)
+            LOG.exception("Replication for the volume %s has "
+                          "failed.", rvol_name)
             self._delete_by_ref(self.target, "replication/sessions",
                                 rsession_name, 'session')
             self._delete_by_ref(self.client, "replication/sessions",
                                 session_name, 'remote session')
             self._delete_by_ref(self.client, "volumes", vol_name, "volume")
             self._delete_by_ref(self.client, "volume_groups", vg_name, "vg")
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
-    @kaminario_logger
+    @utils.trace
     def _delete_by_ref(self, device, url, name, msg):
         rs = device.search(url, name=name)
         for result in rs.hits:
             result.delete()
             LOG.debug("Deleting %(msg)s: %(name)s", {'msg': msg, 'name': name})
 
-    @kaminario_logger
+    @utils.trace
     def _failover_volume(self, volume):
         """Promoting a secondary volume to primary volume."""
         session_name = self.get_session_name(volume.id)
@@ -368,8 +347,8 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             LOG.debug("The target session: %s state is "
                       "changed to failed_over ", rsession_name)
 
-    @kaminario_logger
-    def failover_host(self, context, volumes, secondary_id=None):
+    @utils.trace
+    def failover_host(self, context, volumes, secondary_id=None, groups=None):
         """Failover to replication target."""
         volume_updates = []
         back_end_ip = None
@@ -378,8 +357,8 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                                               'cinder-volume')
 
         if secondary_id and secondary_id != self.replica.backend_id:
-            LOG.error(_LE("Kaminario driver received failover_host "
-                          "request, But backend is non replicated device"))
+            LOG.error("Kaminario driver received failover_host "
+                      "request, But backend is non replicated device")
             raise exception.UnableToFailOver(reason=_("Failover requested "
                                                       "on non replicated "
                                                       "backend."))
@@ -520,9 +499,9 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                     volume_updates.append({'volume_id': v['id'],
                                            'updates': {'status': 'error', }})
             back_end_ip = self.replica.backend_id
-        return back_end_ip, volume_updates
+        return back_end_ip, volume_updates, []
 
-    @kaminario_logger
+    @utils.trace
     def _create_volume_replica_user_snap(self, k2, sess):
         snap = k2.new("snapshots")
         snap.is_application_consistent = "False"
@@ -551,7 +530,7 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                     self.snap_updates.remove(l)
                 eventlet.sleep(1)
 
-    @kaminario_logger
+    @utils.trace
     def create_volume_from_snapshot(self, volume, snapshot):
         """Create volume from snapshot.
 
@@ -579,11 +558,10 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                                         source=snap, retention_policy=rpolicy,
                                         is_exposable=True).save()
             except Exception as ex:
-                LOG.exception(_LE("Creating a view: %(view)s from snapshot: "
-                                  "%(snap)s failed"), {"view": view_name,
-                                                       "snap": snap_name})
-                raise exception.KaminarioCinderDriverException(
-                    reason=six.text_type(ex.message))
+                LOG.exception("Creating a view: %(view)s from snapshot: "
+                              "%(snap)s failed", {"view": view_name,
+                                                  "snap": snap_name})
+                raise exception.KaminarioCinderDriverException(reason=ex)
 
         else:
             msg = _("Snapshot: %s search failed in K2.") % snap_name
@@ -613,12 +591,11 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             self.terminate_connection(volume, properties)
             cview.delete()
             self.delete_volume(volume)
-            LOG.exception(_LE("Copy to volume: %(vol)s from view: %(view)s "
-                              "failed"), {"vol": vol_name, "view": view_name})
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            LOG.exception("Copy to volume: %(vol)s from view: %(view)s "
+                          "failed", {"vol": vol_name, "view": view_name})
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
-    @kaminario_logger
+    @utils.trace
     def create_cloned_volume(self, volume, src_vref):
         """Create a clone from source volume.
 
@@ -633,7 +610,7 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
         src_map = self.client.search("mappings", volume=src_vol)
         src_attach_info = dest_attach_info = None
         if src_map.total != 0:
-            msg = _("K2 driver does not support clone of a attached volume. "
+            msg = _("K2 driver does not support clone of an attached volume. "
                     "To get this done, create a snapshot from the attached "
                     "volume and then create a volume from the snapshot.")
             LOG.error(msg)
@@ -660,11 +637,10 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             self.terminate_connection(src_vref, properties)
             self.terminate_connection(volume, properties)
             self.delete_volume(volume)
-            LOG.exception(_LE("Create a clone: %s failed."), clone_name)
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            LOG.exception("Create a clone: %s failed.", clone_name)
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
-    @kaminario_logger
+    @utils.trace
     def delete_volume(self, volume):
         """Volume in K2 exists in a volume group.
 
@@ -686,11 +662,10 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             if vg_rs.total != 0:
                 vg_rs.hits[0].delete()
         except Exception as ex:
-            LOG.exception(_LE("Deletion of volume %s failed."), vol_name)
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            LOG.exception("Deletion of volume %s failed.", vol_name)
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
-    @kaminario_logger
+    @utils.trace
     def _delete_volume_replica(self, volume, vg_name, vol_name):
         rvg_name = self.get_rep_name(vg_name)
         rvol_name = self.get_rep_name(vol_name)
@@ -724,7 +699,7 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
         self._delete_by_ref(self.target, "volume_groups",
                             rvg_name, "remote vg")
 
-    @kaminario_logger
+    @utils.trace
     def _delete_failover_volume_replica(self, volume, vg_name, vol_name):
         rvg_name = self.get_rep_name(vg_name)
         rvol_name = self.get_rep_name(vol_name)
@@ -749,7 +724,7 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             obj.refresh()
             eventlet.sleep(1)
 
-    @kaminario_logger
+    @utils.trace
     def get_volume_stats(self, refresh=False):
         if refresh:
             self.update_volume_stats()
@@ -771,7 +746,7 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
     def remove_export(self, context, volume):
         pass
 
-    @kaminario_logger
+    @utils.trace
     def create_snapshot(self, snapshot):
         """Create a snapshot from a volume_group."""
         vg_name = self.get_volume_group_name(snapshot.volume_id)
@@ -786,11 +761,10 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                             source=vg, retention_policy=rpolicy,
                             is_auto_deleteable=False).save()
         except Exception as ex:
-            LOG.exception(_LE("Creation of snapshot: %s failed."), snap_name)
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            LOG.exception("Creation of snapshot: %s failed.", snap_name)
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
-    @kaminario_logger
+    @utils.trace
     def delete_snapshot(self, snapshot):
         """Delete a snapshot."""
         snap_name = self.get_snap_name(snapshot.id)
@@ -800,11 +774,10 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             if snap_rs.total != 0:
                 snap_rs.hits[0].delete()
         except Exception as ex:
-            LOG.exception(_LE("Deletion of snapshot: %s failed."), snap_name)
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            LOG.exception("Deletion of snapshot: %s failed.", snap_name)
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
-    @kaminario_logger
+    @utils.trace
     def extend_volume(self, volume, new_size):
         """Extend volume."""
         vol_name = self.get_volume_name(volume.id)
@@ -815,9 +788,8 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             LOG.debug("Extending volume: %s in K2.", vol_name)
             vol.save()
         except Exception as ex:
-            LOG.exception(_LE("Extending volume: %s failed."), vol_name)
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            LOG.exception("Extending volume: %s failed.", vol_name)
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
     def update_volume_stats(self):
         conf = self.configuration
@@ -826,11 +798,14 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
         LOG.debug("Searching total volumes in K2 for updating stats.")
         total_volumes = self.client.search("volumes").total - 1
         provisioned_vol = cap.provisioned_volumes
+
         if (conf.auto_calc_max_oversubscription_ratio and cap.provisioned
                 and (cap.total - cap.free) != 0):
             ratio = provisioned_vol / float(cap.total - cap.free)
         else:
-            ratio = conf.max_over_subscription_ratio
+            ratio = vol_utils.get_max_over_subscription_ratio(
+                conf.max_over_subscription_ratio, supports_auto=True)
+
         self.stats = {'QoS_support': False,
                       'free_capacity_gb': cap.free / units.Mi,
                       'total_capacity_gb': cap.total / units.Mi,
@@ -839,19 +814,26 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                       'total_volumes': total_volumes,
                       'thick_provisioning_support': False,
                       'provisioned_capacity_gb': provisioned_vol / units.Mi,
-                      'max_oversubscription_ratio': ratio,
+                      'max_over_subscription_ratio': ratio,
                       'kaminario:thin_prov_type': 'dedup/nodedup',
                       'replication_enabled': True,
                       'kaminario:replication': True}
 
     def get_initiator_host_name(self, connector):
-        """Return the initiator host name.
+        """Return the initiator host name or unique ID.
+
+        Unique ID when configuration's unique_fqdn_network is false will be
+        the reversed IQN/WWPNS.
 
         Valid characters: 0-9, a-z, A-Z, '-', '_'
         All other characters are replaced with '_'.
         Total characters in initiator host name: 32
         """
-        return re.sub('[^0-9a-zA-Z-_]', '_', connector.get('host', ''))[:32]
+        name = connector.get('initiator',
+                             connector.get('wwnns', [''])[0])[::-1]
+        if self.configuration.unique_fqdn_network:
+            name = connector.get('host', name)
+        return re.sub('[^0-9a-zA-Z-_]', '_', name[:32])
 
     def get_volume_group_name(self, vid):
         """Return the volume group name."""
@@ -877,7 +859,7 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
         """Return the corresponding replication names."""
         return "r{0}".format(name)
 
-    @kaminario_logger
+    @utils.trace
     def _delete_host_by_name(self, name):
         """Deleting host by name."""
         host_rs = self.client.search("hosts", name=name)
@@ -892,9 +874,8 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             return self.client.search("retention_policies",
                                       name="Best_Effort_Retention").hits[0]
         except Exception as ex:
-            LOG.exception(_LE("Retention policy search failed in K2."))
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            LOG.exception("Retention policy search failed in K2.")
+            raise exception.KaminarioCinderDriverException(reason=ex)
 
     def _get_volume_object(self, volume):
         vol_name = self.get_volume_name(volume.id)
@@ -922,7 +903,7 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
     def initialize_connection(self, volume, connector):
         pass
 
-    @kaminario_logger
+    @utils.trace
     def terminate_connection(self, volume, connector):
         """Terminate connection of volume from host."""
         # Get volume object
@@ -936,9 +917,14 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                 volume = volume_rs.hits[0]
         else:
             vol_name = volume.name
-
-        # Get host object.
-        host_name = self.get_initiator_host_name(connector)
+        host_name = ""
+        if connector is None:
+            vol_map_rs = self.client.search("mappings", {"volume": volume})
+            if hasattr(vol_map_rs, "hits") and vol_map_rs.total != 0:
+                host_name = vol_map_rs.hits[0].host.name
+        else:
+            # Get host object.
+            host_name = self.get_initiator_host_name(connector)
         host_rs = self.client.search("hosts", name=host_name)
         if hasattr(host_rs, "hits") and host_rs.total != 0 and volume:
             host = host_rs.hits[0]
@@ -951,9 +937,9 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                 LOG.debug("Deleting initiator hostname: %s in K2.", host_name)
                 host.delete()
         else:
-            LOG.warning(_LW("Host: %s not found on K2."), host_name)
+            LOG.warning("Host: %s not found on K2.", host_name)
 
-    @kaminario_logger
+    @utils.trace
     def k2_initialize_connection(self, volume, connector):
         # Get volume object.
         if type(volume).__name__ != 'RestObject':
@@ -970,11 +956,10 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
         except Exception as ex:
             if host_rs.total == 0:
                 self._delete_host_by_name(host_name)
-            LOG.exception(_LE("Unable to map volume: %(vol)s to host: "
-                              "%(host)s"), {'host': host_name,
-                          'vol': vol.name})
-            raise exception.KaminarioCinderDriverException(
-                reason=six.text_type(ex.message))
+            LOG.exception("Unable to map volume: %(vol)s to host: "
+                          "%(host)s", {'host': host_name,
+                                       'vol': vol.name})
+            raise exception.KaminarioCinderDriverException(reason=ex)
         # Get lun number.
         if type(volume).__name__ == 'RestObject':
             return self._get_lun_number(vol, host)
@@ -990,12 +975,6 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                 'kaminario:thin_prov_type')
             if specs_val == 'nodedup':
                 return False
-            elif CONF.kaminario_nodedup_substring in vol_type.get('name'):
-                LOG.info(_LI("'kaminario_nodedup_substring' option is "
-                             "deprecated in favour of 'kaminario:thin_prov_"
-                             "type' in extra-specs and will be removed in "
-                             "the 10.0.0 release."))
-                return False
             else:
                 return True
         else:
@@ -1019,25 +998,35 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                 return True
         return False
 
-    @kaminario_logger
+    @utils.trace
     def manage_existing(self, volume, existing_ref):
         vol_name = existing_ref['source-name']
         new_name = self.get_volume_name(volume.id)
         vg_new_name = self.get_volume_group_name(volume.id)
         vg_name = None
         is_dedup = self._get_is_dedup(volume.get('volume_type'))
+        reason = None
         try:
             LOG.debug("Searching volume: %s in K2.", vol_name)
             vol = self.client.search("volumes", name=vol_name).hits[0]
             vg = vol.volume_group
+            nvol = self.client.search("volumes", volume_group=vg).total
             vg_replica = self._get_replica_status(vg.name)
             vol_map = False
             if self.client.search("mappings", volume=vol).total != 0:
                 vol_map = True
-            if is_dedup != vg.is_dedup or vg_replica or vol_map:
+            if is_dedup != vg.is_dedup:
+                reason = 'dedup type mismatch for K2 volume group.'
+            elif vg_replica:
+                reason = 'replication enabled K2 volume group.'
+            elif vol_map:
+                reason = 'attached K2 volume.'
+            elif nvol != 1:
+                reason = 'multiple volumes in K2 volume group.'
+            if reason:
                 raise exception.ManageExistingInvalidReference(
                     existing_ref=existing_ref,
-                    reason=_('Manage volume type invalid.'))
+                    reason=_('Unable to manage K2 volume due to: %s') % reason)
             vol.name = new_name
             vg_name = vg.name
             LOG.debug("Manage new volume name: %s", new_name)
@@ -1046,7 +1035,11 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             vg.save()
             LOG.debug("Manage volume: %s in K2.", vol_name)
             vol.save()
-        except Exception as ex:
+        except exception.ManageExistingInvalidReference:
+            LOG.exception("manage volume: %s failed.", vol_name)
+            raise
+        except Exception:
+            LOG.exception("manage volume: %s failed.", vol_name)
             vg_rs = self.client.search("volume_groups", name=vg_new_name)
             if hasattr(vg_rs, 'hits') and vg_rs.total != 0:
                 vg = vg_rs.hits[0]
@@ -1054,12 +1047,9 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                     vg.name = vg_name
                     LOG.debug("Updating vg new name to old name: %s ", vg_name)
                     vg.save()
-            LOG.exception(_LE("manage volume: %s failed."), vol_name)
-            raise exception.ManageExistingInvalidReference(
-                existing_ref=existing_ref,
-                reason=six.text_type(ex.message))
+            raise
 
-    @kaminario_logger
+    @utils.trace
     def manage_existing_get_size(self, volume, existing_ref):
         vol_name = existing_ref['source-name']
         v_rs = self.client.search("volumes", name=vol_name)
@@ -1072,7 +1062,7 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
                 existing_ref=existing_ref,
                 reason=_('Unable to get size of manage volume.'))
 
-    @kaminario_logger
+    @utils.trace
     def after_volume_copy(self, ctxt, volume, new_volume, remote=None):
         self.delete_volume(volume)
         vg_name_old = self.get_volume_group_name(volume.id)
@@ -1086,7 +1076,7 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
         vol_new.name = vol_name_old
         vol_new.save()
 
-    @kaminario_logger
+    @utils.trace
     def retype(self, ctxt, volume, new_type, diff, host):
         old_type = volume.get('volume_type')
         vg_name = self.get_volume_group_name(volume.id)
@@ -1116,8 +1106,8 @@ class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
             LOG.debug(msg)
             return False
         else:
-            LOG.error(_LE('Change from type1: %(type1)s to type2: %(type2)s '
-                          'is not supported directly in K2.'),
+            LOG.error('Change from type1: %(type1)s to type2: %(type2)s '
+                      'is not supported directly in K2.',
                       {'type1': old_type, 'type2': new_type})
             return False
 
