@@ -10,10 +10,11 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import collections
+import six
 
 from oslo_config import cfg
 from oslo_log import log as logging
-from oslo_utils import timeutils
 from oslo_utils import units
 import taskflow.engines
 from taskflow.patterns import linear_flow
@@ -21,12 +22,15 @@ from taskflow.types import failure as ft
 
 from cinder import exception
 from cinder import flow_utils
-from cinder.i18n import _, _LE, _LW
+from cinder.i18n import _
 from cinder import objects
-from cinder import policy
+from cinder.objects import fields
+from cinder.policies import volumes as policy
 from cinder import quota
+from cinder import quota_utils
 from cinder import utils
 from cinder.volume.flows import common
+from cinder.volume import utils as vol_utils
 from cinder.volume import volume_types
 
 LOG = logging.getLogger(__name__)
@@ -39,11 +43,13 @@ QUOTAS = quota.QUOTAS
 # Only in these 'sources' status can we attempt to create a volume from a
 # source volume or a source snapshot, other status states we can not create
 # from, 'error' being the common example.
-SNAPSHOT_PROCEED_STATUS = ('available',)
+SNAPSHOT_PROCEED_STATUS = (fields.SnapshotStatus.AVAILABLE,)
 SRC_VOL_PROCEED_STATUS = ('available', 'in-use',)
 REPLICA_PROCEED_STATUS = ('active', 'active-stopped',)
 CG_PROCEED_STATUS = ('available', 'creating',)
 CGSNAPSHOT_PROCEED_STATUS = ('available',)
+GROUP_PROCEED_STATUS = ('available', 'creating',)
+BACKUP_PROCEED_STATUS = (fields.BackupStatus.AVAILABLE,)
 
 
 class ExtractVolumeRequestTask(flow_utils.CinderTask):
@@ -60,11 +66,12 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
     # This task will produce the following outputs (said outputs can be
     # saved to durable storage in the future so that the flow can be
     # reconstructed elsewhere and continued).
-    default_provides = set(['availability_zone', 'size', 'snapshot_id',
+    default_provides = set(['size', 'snapshot_id',
                             'source_volid', 'volume_type', 'volume_type_id',
-                            'encryption_key_id', 'source_replicaid',
-                            'consistencygroup_id', 'cgsnapshot_id',
-                            'qos_specs'])
+                            'encryption_key_id', 'consistencygroup_id',
+                            'cgsnapshot_id', 'qos_specs', 'group_id',
+                            'refresh_az', 'backup_id', 'availability_zones',
+                            'multiattach'])
 
     def __init__(self, image_service, availability_zones, **kwargs):
         super(ExtractVolumeRequestTask, self).__init__(addons=[ACTION],
@@ -112,6 +119,11 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
                                       exception.InvalidConsistencyGroup,
                                       'consistencygroup')
 
+    def _extract_group(self, group):
+        return self._extract_resource(group, (GROUP_PROCEED_STATUS,),
+                                      exception.InvalidGroup,
+                                      'group')
+
     def _extract_cgsnapshot(self, cgsnapshot):
         return self._extract_resource(cgsnapshot, (CGSNAPSHOT_PROCEED_STATUS,),
                                       exception.InvalidCgSnapshot,
@@ -125,15 +137,13 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
         return self._extract_resource(source_volume, (SRC_VOL_PROCEED_STATUS,),
                                       exception.InvalidVolume, 'source volume')
 
-    def _extract_source_replica(self, source_replica):
-        return self._extract_resource(source_replica, (SRC_VOL_PROCEED_STATUS,
-                                                       REPLICA_PROCEED_STATUS),
-                                      exception.InvalidVolume,
-                                      'replica', ('status',
-                                                  'replication_status'))
+    def _extract_backup(self, backup):
+        return self._extract_resource(backup, (BACKUP_PROCEED_STATUS,),
+                                      exception.InvalidBackup,
+                                      'backup')
 
     @staticmethod
-    def _extract_size(size, source_volume, snapshot):
+    def _extract_size(size, source_volume, snapshot, backup):
         """Extracts and validates the volume size.
 
         This function will validate or when not provided fill in the provided
@@ -159,8 +169,17 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
                              'source_size': source_volume['size']}
                 raise exception.InvalidInput(reason=msg)
 
+        def validate_backup_size(size):
+            if backup and size < backup['size']:
+                msg = _("Volume size %(size)sGB cannot be smaller than "
+                        "the backup size %(backup_size)sGB. "
+                        "It must be >= backup size.")
+                msg = msg % {'size': size,
+                             'backup_size': backup['size']}
+                raise exception.InvalidInput(reason=msg)
+
         def validate_int(size):
-            if not isinstance(size, int) or size <= 0:
+            if not isinstance(size, six.integer_types) or size <= 0:
                 msg = _("Volume size '%(size)s' must be an integer and"
                         " greater than 0") % {'size': size}
                 raise exception.InvalidInput(reason=msg)
@@ -172,15 +191,19 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
             validator_functors.append(validate_source_size)
         elif snapshot:
             validator_functors.append(validate_snap_size)
+        elif backup:
+            validator_functors.append(validate_backup_size)
 
         # If the size is not provided then try to provide it.
         if not size and source_volume:
             size = source_volume['size']
         elif not size and snapshot:
             size = snapshot.volume_size
+        elif not size and backup:
+            size = backup['size']
 
         size = utils.as_int(size)
-        LOG.debug("Validating volume '%(size)s' using %(functors)s" %
+        LOG.debug("Validating volume size '%(size)s' using %(functors)s",
                   {'size': size,
                    'functors': ", ".join([common.make_pretty_name(func)
                                           for func in validator_functors])})
@@ -188,12 +211,15 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
             func(size)
         return size
 
-    def _check_image_metadata(self, context, image_id, size):
-        """Checks image existence and validates that the image metadata."""
+    def _get_image_metadata(self, context, image_id, size):
+        """Checks image existence and validates the image metadata.
+
+        Returns: image metadata or None
+        """
 
         # Check image existence
         if image_id is None:
-            return
+            return None
 
         # NOTE(harlowja): this should raise an error if the image does not
         # exist, this is expected as it signals that the image_id is missing.
@@ -207,7 +233,7 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
 
         # Check image size is not larger than volume size.
         image_size = utils.as_int(image_meta['size'], quiet=False)
-        image_size_in_gb = (image_size + GB - 1) / GB
+        image_size_in_gb = (image_size + GB - 1) // GB
         if image_size_in_gb > size:
             msg = _('Size of specified image %(image_size)sGB'
                     ' is larger than volume size %(volume_size)sGB.')
@@ -222,38 +248,36 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
             msg = msg % {'volume_size': size, 'min_disk': min_disk}
             raise exception.InvalidInput(reason=msg)
 
-    @staticmethod
-    def _check_metadata_properties(metadata=None):
-        """Checks that the volume metadata properties are valid."""
+        return image_meta
 
-        if not metadata:
-            metadata = {}
-
-        for (k, v) in metadata.items():
-            if len(k) == 0:
-                msg = _("Metadata property key blank")
-                LOG.warning(msg)
-                raise exception.InvalidVolumeMetadata(reason=msg)
-            if len(k) > 255:
-                msg = _("Metadata property key %s greater than 255 "
-                        "characters") % k
-                LOG.warning(msg)
-                raise exception.InvalidVolumeMetadataSize(reason=msg)
-            if len(v) > 255:
-                msg = _("Metadata property key %s value greater than"
-                        " 255 characters") % k
-                LOG.warning(msg)
-                raise exception.InvalidVolumeMetadataSize(reason=msg)
-
-    def _extract_availability_zone(self, availability_zone, snapshot,
-                                   source_volume):
-        """Extracts and returns a validated availability zone.
+    def _extract_availability_zones(self, availability_zone, snapshot,
+                                    source_volume, group, volume_type=None):
+        """Extracts and returns a validated availability zone list.
 
         This function will extract the availability zone (if not provided) from
         the snapshot or source_volume and then performs a set of validation
         checks on the provided or extracted availability zone and then returns
         the validated availability zone.
         """
+        refresh_az = False
+        type_azs = vol_utils.extract_availability_zones_from_volume_type(
+            volume_type)
+        type_az_configured = type_azs is not None
+        if type_az_configured:
+            safe_azs = list(
+                set(type_azs).intersection(self.availability_zones))
+            if not safe_azs:
+                raise exception.InvalidTypeAvailabilityZones(az=type_azs)
+        else:
+            safe_azs = self.availability_zones
+
+        # If the volume will be created in a group, it should be placed in
+        # in same availability zone as the group.
+        if group:
+            try:
+                availability_zone = group['availability_zone']
+            except (TypeError, KeyError):
+                pass
 
         # Try to extract the availability zone from the corresponding snapshot
         # or source volume if either is valid so that we can be in the same
@@ -270,32 +294,31 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
                 except (TypeError, KeyError):
                     pass
 
-        if availability_zone is None:
+        if availability_zone is None and not type_az_configured:
             if CONF.default_availability_zone:
                 availability_zone = CONF.default_availability_zone
             else:
                 # For backwards compatibility use the storage_availability_zone
                 availability_zone = CONF.storage_availability_zone
 
-        if availability_zone not in self.availability_zones:
+        if availability_zone and availability_zone not in safe_azs:
+            refresh_az = True
             if CONF.allow_availability_zone_fallback:
                 original_az = availability_zone
                 availability_zone = (
                     CONF.default_availability_zone or
                     CONF.storage_availability_zone)
-                LOG.warning(_LW("Availability zone '%(s_az)s' "
-                                "not found, falling back to "
-                                "'%(s_fallback_az)s'."),
+                LOG.warning("Availability zone '%(s_az)s' "
+                            "not found, falling back to "
+                            "'%(s_fallback_az)s'.",
                             {'s_az': original_az,
                              's_fallback_az': availability_zone})
             else:
-                msg = _("Availability zone '%(s_az)s' is invalid.")
-                msg = msg % {'s_az': availability_zone}
-                raise exception.InvalidInput(reason=msg)
+                raise exception.InvalidAvailabilityZone(az=availability_zone)
 
         # If the configuration only allows cloning to the same availability
         # zone then we need to enforce that.
-        if CONF.cloned_volume_same_az:
+        if availability_zone and CONF.cloned_volume_same_az:
             snap_az = None
             try:
                 snap_az = snapshot['volume']['availability_zone']
@@ -315,121 +338,168 @@ class ExtractVolumeRequestTask(flow_utils.CinderTask):
                         "availability zone as the source volume")
                 raise exception.InvalidInput(reason=msg)
 
-        return availability_zone
+        if availability_zone:
+            return [availability_zone], refresh_az
+        else:
+            return safe_azs, refresh_az
 
     def _get_encryption_key_id(self, key_manager, context, volume_type_id,
-                               snapshot, source_volume):
+                               snapshot, source_volume,
+                               image_metadata):
         encryption_key_id = None
         if volume_types.is_encrypted(context, volume_type_id):
             if snapshot is not None:  # creating from snapshot
                 encryption_key_id = snapshot['encryption_key_id']
             elif source_volume is not None:  # cloning volume
                 encryption_key_id = source_volume['encryption_key_id']
+            elif image_metadata is not None:
+                # creating from image
+                encryption_key_id = image_metadata.get(
+                    'cinder_encryption_key_id')
 
             # NOTE(joel-coffman): References to the encryption key should *not*
             # be copied because the key is deleted when the volume is deleted.
             # Clone the existing key and associate a separate -- but
             # identical -- key with each volume.
             if encryption_key_id is not None:
-                encryption_key_id = key_manager.copy_key(context,
-                                                         encryption_key_id)
+                encryption_key_id = vol_utils.clone_encryption_key(
+                    context,
+                    key_manager,
+                    encryption_key_id)
             else:
-                encryption_key_id = key_manager.create_key(context)
+                encryption_key_id = vol_utils.create_encryption_key(
+                    context,
+                    key_manager,
+                    volume_type_id)
 
         return encryption_key_id
 
-    def _get_volume_type_id(self, volume_type, source_volume, snapshot):
-        if not volume_type and source_volume:
-            return source_volume['volume_type_id']
-        elif snapshot is not None:
-            if volume_type:
-                current_volume_type_id = volume_type.get('id')
-                if current_volume_type_id != snapshot['volume_type_id']:
-                    msg = _LW("Volume type will be changed to "
-                              "be the same as the source volume.")
-                    LOG.warning(msg)
-            return snapshot['volume_type_id']
-        else:
-            return volume_type.get('id')
+    def _get_volume_type(self, context, volume_type,
+                         source_volume, snapshot, image_volume_type_id):
+        if volume_type:
+            return volume_type
+        identifier = collections.defaultdict(str)
+        try:
+            if source_volume:
+                identifier = {'source': 'volume',
+                              'id': source_volume['volume_type_id']}
+            elif snapshot:
+                identifier = {'source': 'snapshot',
+                              'id': snapshot['volume_type_id']}
+            elif image_volume_type_id:
+                identifier = {'source': 'image',
+                              'id': image_volume_type_id}
+            elif CONF.default_volume_type:
+                identifier = {'source': 'default volume type config',
+                              'id': CONF.default_volume_type}
+            if identifier:
+                return objects.VolumeType.get_by_name_or_id(
+                    context, identifier['id'])
+        except (exception.VolumeTypeNotFound,
+                exception.VolumeTypeNotFoundByName,
+                exception.InvalidVolumeType):
+            LOG.exception("Failed to find volume type from "
+                          "source %(source)s, identifier %(id)s", identifier)
+        return None
 
     def execute(self, context, size, snapshot, image_id, source_volume,
                 availability_zone, volume_type, metadata, key_manager,
-                source_replica, consistencygroup, cgsnapshot):
+                consistencygroup, cgsnapshot, group, group_snapshot, backup,
+                multiattach=False):
 
         utils.check_exclusive_options(snapshot=snapshot,
                                       imageRef=image_id,
-                                      source_volume=source_volume)
-        policy.enforce_action(context, ACTION)
+                                      source_volume=source_volume,
+                                      backup=backup)
+        context.authorize(policy.CREATE_POLICY)
 
         # TODO(harlowja): what guarantee is there that the snapshot or source
         # volume will remain available after we do this initial verification??
         snapshot_id = self._extract_snapshot(snapshot)
         source_volid = self._extract_source_volume(source_volume)
-        source_replicaid = self._extract_source_replica(source_replica)
-        size = self._extract_size(size, source_volume, snapshot)
+        backup_id = self._extract_backup(backup)
+        size = self._extract_size(size, source_volume, snapshot, backup)
         consistencygroup_id = self._extract_consistencygroup(consistencygroup)
         cgsnapshot_id = self._extract_cgsnapshot(cgsnapshot)
+        group_id = self._extract_group(group)
 
-        self._check_image_metadata(context, image_id, size)
+        image_meta = self._get_image_metadata(context,
+                                              image_id,
+                                              size)
 
-        availability_zone = self._extract_availability_zone(availability_zone,
-                                                            snapshot,
-                                                            source_volume)
+        image_properties = image_meta.get(
+            'properties', {}) if image_meta else {}
+        image_volume_type = image_properties.get(
+            'cinder_img_volume_type', None) if image_properties else None
 
-        # TODO(joel-coffman): This special handling of snapshots to ensure that
-        # their volume type matches the source volume is too convoluted. We
-        # should copy encryption metadata from the encrypted volume type to the
-        # volume upon creation and propagate that information to each snapshot.
-        # This strategy avoids any dependency upon the encrypted volume type.
-        def_vol_type = volume_types.get_default_volume_type()
-        if not volume_type and not source_volume and not snapshot:
-            volume_type = def_vol_type
+        volume_type = self._get_volume_type(
+            context, volume_type, source_volume, snapshot, image_volume_type)
 
-        # When creating a clone of a replica (replication test), we can't
-        # use the volume type of the replica, therefore, we use the default.
-        # NOTE(ronenkat): this assumes the default type is not replicated.
-        if source_replicaid:
-            volume_type = def_vol_type
+        volume_type_id = volume_type.get('id') if volume_type else None
 
-        volume_type_id = self._get_volume_type_id(volume_type,
-                                                  source_volume, snapshot)
+        availability_zones, refresh_az = self._extract_availability_zones(
+            availability_zone, snapshot, source_volume, group,
+            volume_type=volume_type)
 
-        if image_id and volume_types.is_encrypted(context, volume_type_id):
-            msg = _('Create encrypted volumes with type %(type)s '
-                    'from image %(image)s is not supported.')
-            msg = msg % {'type': volume_type_id,
-                         'image': image_id, }
-            raise exception.InvalidInput(reason=msg)
+        encryption_key_id = self._get_encryption_key_id(
+            key_manager,
+            context,
+            volume_type_id,
+            snapshot,
+            source_volume,
+            image_meta)
 
-        encryption_key_id = self._get_encryption_key_id(key_manager,
-                                                        context,
-                                                        volume_type_id,
-                                                        snapshot,
-                                                        source_volume)
+        if volume_type_id:
+            volume_type = objects.VolumeType.get_by_name_or_id(
+                context, volume_type_id)
+            extra_specs = volume_type.get('extra_specs', {})
+            # NOTE(tommylikehu): Although the parameter `multiattach` from
+            # create volume API is deprecated now, we still need to consider
+            # it when multiattach is not enabled in volume type.
+            multiattach = (extra_specs.get(
+                'multiattach', '') == '<is> True' or multiattach)
+            if multiattach and encryption_key_id:
+                msg = _('Multiattach cannot be used with encrypted volumes.')
+                raise exception.InvalidVolume(reason=msg)
+
+        if multiattach:
+            context.authorize(policy.MULTIATTACH_POLICY)
 
         specs = {}
         if volume_type_id:
             qos_specs = volume_types.get_volume_type_qos_specs(volume_type_id)
             if qos_specs['qos_specs']:
                 specs = qos_specs['qos_specs'].get('specs', {})
+
+            # Determine default replication status
+            extra_specs = volume_types.get_volume_type_extra_specs(
+                volume_type_id)
         if not specs:
             # to make sure we don't pass empty dict
             specs = None
+            extra_specs = None
 
-        self._check_metadata_properties(metadata)
+        if vol_utils.is_replicated_spec(extra_specs):
+            replication_status = fields.ReplicationStatus.ENABLED
+        else:
+            replication_status = fields.ReplicationStatus.DISABLED
 
         return {
             'size': size,
             'snapshot_id': snapshot_id,
             'source_volid': source_volid,
-            'availability_zone': availability_zone,
             'volume_type': volume_type,
             'volume_type_id': volume_type_id,
             'encryption_key_id': encryption_key_id,
             'qos_specs': specs,
-            'source_replicaid': source_replicaid,
             'consistencygroup_id': consistencygroup_id,
             'cgsnapshot_id': cgsnapshot_id,
+            'group_id': group_id,
+            'replication_status': replication_status,
+            'refresh_az': refresh_az,
+            'backup_id': backup_id,
+            'multiattach': multiattach,
+            'availability_zones': availability_zones
         }
 
 
@@ -441,15 +511,14 @@ class EntryCreateTask(flow_utils.CinderTask):
 
     default_provides = set(['volume_properties', 'volume_id', 'volume'])
 
-    def __init__(self, db):
-        requires = ['availability_zone', 'description', 'metadata',
+    def __init__(self):
+        requires = ['description', 'metadata',
                     'name', 'reservations', 'size', 'snapshot_id',
                     'source_volid', 'volume_type_id', 'encryption_key_id',
-                    'source_replicaid', 'consistencygroup_id',
-                    'cgsnapshot_id', 'multiattach', 'qos_specs']
+                    'consistencygroup_id', 'cgsnapshot_id', 'multiattach',
+                    'qos_specs', 'group_id', 'availability_zones']
         super(EntryCreateTask, self).__init__(addons=[ACTION],
                                               requires=requires)
-        self.db = db
 
     def execute(self, context, optional_args, **kwargs):
         """Creates a database entry for the given inputs and returns details.
@@ -461,24 +530,52 @@ class EntryCreateTask(flow_utils.CinderTask):
         pre-cursor task.
         """
 
+        src_volid = kwargs.get('source_volid')
+        src_vol = None
+        if src_volid is not None:
+            src_vol = objects.Volume.get_by_id(context, src_volid)
+        bootable = False
+        if src_vol is not None:
+            bootable = src_vol.bootable
+        elif kwargs.get('snapshot_id'):
+            snapshot = objects.Snapshot.get_by_id(context,
+                                                  kwargs.get('snapshot_id'))
+            volume_id = snapshot.volume_id
+            snp_vol = objects.Volume.get_by_id(context, volume_id)
+            if snp_vol is not None:
+                bootable = snp_vol.bootable
+        availability_zones = kwargs.pop('availability_zones')
         volume_properties = {
             'size': kwargs.pop('size'),
             'user_id': context.user_id,
             'project_id': context.project_id,
             'status': 'creating',
-            'attach_status': 'detached',
+            'attach_status': fields.VolumeAttachStatus.DETACHED,
             'encryption_key_id': kwargs.pop('encryption_key_id'),
             # Rename these to the internal name.
             'display_description': kwargs.pop('description'),
             'display_name': kwargs.pop('name'),
-            'replication_status': 'disabled',
             'multiattach': kwargs.pop('multiattach'),
+            'bootable': bootable,
         }
+        if len(availability_zones) == 1:
+            volume_properties['availability_zone'] = availability_zones[0]
 
         # Merge in the other required arguments which should provide the rest
         # of the volume property fields (if applicable).
         volume_properties.update(kwargs)
-        volume = self.db.volume_create(context, volume_properties)
+        volume = objects.Volume(context=context, **volume_properties)
+        volume.create()
+
+        # FIXME(dulek): We're passing this volume_properties dict through RPC
+        # in request_spec. This shouldn't be needed, most data is replicated
+        # in both volume and other places. We should make Newton read data
+        # from just one correct place and leave just compatibility code.
+        #
+        # Right now - let's move it to versioned objects to be able to make
+        # non-backward compatible changes.
+
+        volume_properties = objects.VolumeProperties(**volume_properties)
 
         return {
             'volume_id': volume['id'],
@@ -500,20 +597,20 @@ class EntryCreateTask(flow_utils.CinderTask):
             return
 
         if optional_args['is_quota_committed']:
-            # If quota got commited we shouldn't rollback as the volume has
+            # If quota got committed we shouldn't rollback as the volume has
             # already been created and the quota has already been absorbed.
             return
 
-        vol_id = result['volume_id']
+        volume = result['volume']
         try:
-            self.db.volume_destroy(context.elevated(), vol_id)
+            volume.destroy()
         except exception.CinderException:
             # We are already reverting, therefore we should silence this
             # exception since a second exception being active will be bad.
             #
             # NOTE(harlowja): Being unable to destroy a volume is pretty
             # bad though!!
-            LOG.exception(_LE("Failed destroying volume entry %s"), vol_id)
+            LOG.exception("Failed destroying volume entry %s", volume.id)
 
 
 class QuotaReserveTask(flow_utils.CinderTask):
@@ -535,7 +632,8 @@ class QuotaReserveTask(flow_utils.CinderTask):
     def __init__(self):
         super(QuotaReserveTask, self).__init__(addons=[ACTION])
 
-    def execute(self, context, size, volume_type_id, optional_args):
+    def execute(self, context, size, volume_type_id, group_snapshot,
+                optional_args):
         try:
             values = {'per_volume_gigabytes': size}
             QUOTAS.limit_check(context, project_id=context.project_id,
@@ -546,56 +644,22 @@ class QuotaReserveTask(flow_utils.CinderTask):
                 size=size, limit=quotas['per_volume_gigabytes'])
 
         try:
-            reserve_opts = {'volumes': 1, 'gigabytes': size}
+            if group_snapshot:
+                reserve_opts = {'volumes': 1}
+            else:
+                reserve_opts = {'volumes': 1, 'gigabytes': size}
+            if ('update_size' in optional_args
+                    and optional_args['update_size']):
+                reserve_opts.pop('volumes', None)
             QUOTAS.add_volume_type_opts(context, reserve_opts, volume_type_id)
             reservations = QUOTAS.reserve(context, **reserve_opts)
             return {
                 'reservations': reservations,
             }
         except exception.OverQuota as e:
-            overs = e.kwargs['overs']
-            quotas = e.kwargs['quotas']
-            usages = e.kwargs['usages']
-
-            def _consumed(name):
-                return usages[name]['reserved'] + usages[name]['in_use']
-
-            def _get_over(name):
-                for over in overs:
-                    if name in over:
-                        return over
-                return None
-
-            over_name = _get_over('gigabytes')
-            exceeded_vol_limit_name = _get_over('volumes')
-            if over_name:
-                msg = _LW("Quota exceeded for %(s_pid)s, tried to create "
-                          "%(s_size)sG volume (%(d_consumed)dG "
-                          "of %(d_quota)dG already consumed)")
-                LOG.warning(msg, {'s_pid': context.project_id,
-                                  's_size': size,
-                                  'd_consumed': _consumed(over_name),
-                                  'd_quota': quotas[over_name]})
-                raise exception.VolumeSizeExceedsAvailableQuota(
-                    name=over_name,
-                    requested=size,
-                    consumed=_consumed(over_name),
-                    quota=quotas[over_name])
-            elif exceeded_vol_limit_name:
-                msg = _LW("Quota %(s_name)s exceeded for %(s_pid)s, tried "
-                          "to create volume (%(d_consumed)d volume(s) "
-                          "already consumed).")
-                LOG.warning(msg,
-                            {'s_name': exceeded_vol_limit_name,
-                             's_pid': context.project_id,
-                             'd_consumed':
-                             _consumed(exceeded_vol_limit_name)})
-                raise exception.VolumeLimitExceeded(
-                    allowed=quotas[exceeded_vol_limit_name],
-                    name=exceeded_vol_limit_name)
-            else:
-                # If nothing was reraised, ensure we reraise the initial error
-                raise
+            quota_utils.process_reserve_over_quota(context, e,
+                                                   resource='volumes',
+                                                   size=size)
 
     def revert(self, context, result, optional_args, **kwargs):
         # We never produced a result and therefore can't destroy anything.
@@ -614,8 +678,8 @@ class QuotaReserveTask(flow_utils.CinderTask):
         except exception.CinderException:
             # We are already reverting, therefore we should silence this
             # exception since a second exception being active will be bad.
-            LOG.exception(_LE("Failed rolling back quota for"
-                              " %s reservations"), reservations)
+            LOG.exception("Failed rolling back quota for"
+                          " %s reservations", reservations)
 
 
 class QuotaCommitTask(flow_utils.CinderTask):
@@ -661,8 +725,8 @@ class QuotaCommitTask(flow_utils.CinderTask):
                 QUOTAS.commit(context, reservations,
                               project_id=context.project_id)
         except Exception:
-            LOG.exception(_LE("Failed to update quota for deleting "
-                              "volume: %s"), volume['id'])
+            LOG.exception("Failed to update quota for deleting "
+                          "volume: %s", volume['id'])
 
 
 class VolumeCastTask(flow_utils.CinderTask):
@@ -677,9 +741,10 @@ class VolumeCastTask(flow_utils.CinderTask):
 
     def __init__(self, scheduler_rpcapi, volume_rpcapi, db):
         requires = ['image_id', 'scheduler_hints', 'snapshot_id',
-                    'source_volid', 'volume_id', 'volume_type',
-                    'volume_properties', 'source_replicaid',
-                    'consistencygroup_id', 'cgsnapshot_id', ]
+                    'source_volid', 'volume_id', 'volume', 'volume_type',
+                    'volume_properties', 'consistencygroup_id',
+                    'cgsnapshot_id', 'group_id', 'backup_id',
+                    'availability_zones']
         super(VolumeCastTask, self).__init__(addons=[ACTION],
                                              requires=requires)
         self.volume_rpcapi = volume_rpcapi
@@ -688,17 +753,24 @@ class VolumeCastTask(flow_utils.CinderTask):
 
     def _cast_create_volume(self, context, request_spec, filter_properties):
         source_volid = request_spec['source_volid']
-        source_replicaid = request_spec['source_replicaid']
-        volume_id = request_spec['volume_id']
+        volume = request_spec['volume']
         snapshot_id = request_spec['snapshot_id']
         image_id = request_spec['image_id']
         cgroup_id = request_spec['consistencygroup_id']
-        host = None
-        cgsnapshot_id = request_spec['cgsnapshot_id']
-
+        group_id = request_spec['group_id']
+        backup_id = request_spec['backup_id']
         if cgroup_id:
+            # If cgroup_id existed, we should cast volume to the scheduler
+            # to choose a proper pool whose backend is same as CG's backend.
             cgroup = objects.ConsistencyGroup.get_by_id(context, cgroup_id)
-            host = cgroup.host
+            request_spec['resource_backend'] = vol_utils.extract_host(
+                cgroup.host)
+        elif group_id:
+            # If group_id exists, we should cast volume to the scheduler
+            # to choose a proper pool whose backend is same as group's backend.
+            group = objects.Group.get_by_id(context, group_id)
+            request_spec['resource_backend'] = vol_utils.extract_host(
+                group.host)
         elif snapshot_id and CONF.snapshot_same_host:
             # NOTE(Rongze Zhu): A simple solution for bug 1008866.
             #
@@ -706,63 +778,50 @@ class VolumeCastTask(flow_utils.CinderTask):
             # the call create volume directly to the volume host where the
             # snapshot resides instead of passing it through the scheduler, so
             # snapshot can be copied to the new volume.
+            # NOTE(tommylikehu): In order to check the backend's capacity
+            # before creating volume, we schedule this request to scheduler
+            # service with the desired backend information.
             snapshot = objects.Snapshot.get_by_id(context, snapshot_id)
-            source_volume_ref = self.db.volume_get(context, snapshot.volume_id)
-            host = source_volume_ref['host']
+            request_spec['resource_backend'] = snapshot.volume.host
         elif source_volid:
-            source_volume_ref = self.db.volume_get(context, source_volid)
-            host = source_volume_ref['host']
-        elif source_replicaid:
-            source_volume_ref = self.db.volume_get(context, source_replicaid)
-            host = source_volume_ref['host']
+            source_volume_ref = objects.Volume.get_by_id(context, source_volid)
+            request_spec['resource_backend'] = source_volume_ref.host
 
-        if not host:
-            # Cast to the scheduler and let it handle whatever is needed
-            # to select the target host for this volume.
-            self.scheduler_rpcapi.create_volume(
-                context,
-                CONF.volume_topic,
-                volume_id,
-                snapshot_id=snapshot_id,
-                image_id=image_id,
-                request_spec=request_spec,
-                filter_properties=filter_properties)
-        else:
-            # Bypass the scheduler and send the request directly to the volume
-            # manager.
-            now = timeutils.utcnow()
-            values = {'host': host, 'scheduled_at': now}
-            volume_ref = self.db.volume_update(context, volume_id, values)
-            if not cgsnapshot_id:
-                self.volume_rpcapi.create_volume(
-                    context,
-                    volume_ref,
-                    volume_ref['host'],
-                    request_spec,
-                    filter_properties,
-                    allow_reschedule=False)
+        self.scheduler_rpcapi.create_volume(
+            context,
+            volume,
+            snapshot_id=snapshot_id,
+            image_id=image_id,
+            request_spec=request_spec,
+            filter_properties=filter_properties,
+            backup_id=backup_id)
 
     def execute(self, context, **kwargs):
         scheduler_hints = kwargs.pop('scheduler_hints', None)
-        request_spec = kwargs.copy()
+        db_vt = kwargs.pop('volume_type')
+        kwargs['volume_type'] = None
+        if db_vt:
+            kwargs['volume_type'] = objects.VolumeType()
+            objects.VolumeType()._from_db_object(context,
+                                                 kwargs['volume_type'], db_vt)
+        request_spec = objects.RequestSpec(**kwargs)
         filter_properties = {}
         if scheduler_hints:
             filter_properties['scheduler_hints'] = scheduler_hints
         self._cast_create_volume(context, request_spec, filter_properties)
 
-    def revert(self, context, result, flow_failures, **kwargs):
+    def revert(self, context, result, flow_failures, volume, **kwargs):
         if isinstance(result, ft.Failure):
             return
 
         # Restore the source volume status and set the volume to error status.
-        volume_id = kwargs['volume_id']
         common.restore_source_status(context, self.db, kwargs)
-        common.error_out_volume(context, self.db, volume_id)
-        LOG.error(_LE("Volume %s: create failed"), volume_id)
+        common.error_out(volume)
+        LOG.error("Volume %s: create failed", volume.id)
         exc_info = False
         if all(flow_failures[-1].exc_info):
             exc_info = flow_failures[-1].exc_info
-        LOG.error(_LE('Unexpected build error:'), exc_info=exc_info)
+        LOG.error('Unexpected build error:', exc_info=exc_info)
 
 
 def get_flow(db_api, image_service_api, availability_zones, create_what,
@@ -787,9 +846,10 @@ def get_flow(db_api, image_service_api, availability_zones, create_what,
         availability_zones,
         rebind={'size': 'raw_size',
                 'availability_zone': 'raw_availability_zone',
-                'volume_type': 'raw_volume_type'}))
+                'volume_type': 'raw_volume_type',
+                'multiattach': 'raw_multiattach'}))
     api_flow.add(QuotaReserveTask(),
-                 EntryCreateTask(db_api),
+                 EntryCreateTask(),
                  QuotaCommitTask())
 
     if scheduler_rpcapi and volume_rpcapi:
